"""
The experiment execution API.

Via the function :func:`run_experiment`, you can execute a complex experiment
where multiple optimization algorithms are applied to multiple problem
instances, where log files with the results and progress information about the
runs are collected, and where multiprocessing is used to parallelize the
experiment execution.
Experiments are replicable, as random seeds are automatically generated based
on problem instance names in a replicable fashion.

The log files are structured according to the documentation in
https://thomasweise.github.io/moptipy/#file-names-and-folder-structure
and their contents follow the specification given in
https://thomasweise.github.io/moptipy/#log-file-sections.
"""
import copy
import gc
import os.path
from os import getpid
from typing import Any, Callable, Final, Iterable, Sequence, cast

from numpy.random import Generator, default_rng
from pycommons.ds.cache import str_is_new
from pycommons.io.console import logger
from pycommons.io.path import Path
from pycommons.types import check_int_range, type_error

from moptipy.api.execution import Execution
from moptipy.api.logging import FILE_SUFFIX
from moptipy.api.process import Process
from moptipy.utils.nputils import rand_seeds_from_str
from moptipy.utils.strings import sanitize_name, sanitize_names
from moptipy.utils.sys_info import get_sys_info


def __run_experiment(base_dir: Path,
                     experiments: list[list[Callable]],
                     n_runs: list[int],
                     thread_id: str,
                     perform_warmup: bool,
                     warmup_fes: int,
                     perform_pre_warmup: bool,
                     pre_warmup_fes: int,
                     on_completion: Callable[[
                         Any, Path, Process], None]) -> None:
    """
    Execute a single thread of experiments.

    :param base_dir: the base directory
    :param experiments: the stream of experiment setups
    :param n_runs: the list of runs
    :param thread_id: the thread id
    :param perform_warmup: should we perform a warm-up per instance?
    :param warmup_fes: the number of the FEs for the warm-up runs
    :param perform_pre_warmup: should we do one warmup run for each
        instance before we begin with the actual experiments?
    :param pre_warmup_fes: the FEs for the pre-warmup runs
    :param on_completion: a function to be called for every completed run,
        receiving the instance, the path to the log file (before it is
        created) and the :class:`~moptipy.api.process.Process` of the run
        as parameters
    """
    random: Final[Generator] = default_rng()
    cache: Final[Callable[[str], bool]] = str_is_new()
    for warmup in ([True, False] if perform_pre_warmup else [False]):
        wss: str
        if warmup:
            wss = "pre-warmup"
        else:
            wss = "warmup"
            if perform_pre_warmup:
                gc.collect()  # do full garbage collection after pre-warmups
                gc.collect()  # one more, to be double-safe
                gc.freeze()  # whatever survived now, keep it permanently

        for runs in ([1] if warmup else n_runs):  # for each number of runs
            if not warmup:
                logger(f"now doing {runs} runs.", thread_id)
            random.shuffle(cast(Sequence, experiments))  # shuffle experiments

            for setup in experiments:  # for each setup
                instance = setup[0]()  # load instance
                if instance is None:
                    raise TypeError("None is not an instance.")
                inst_name = sanitize_name(str(instance))

                exp = setup[1](instance)  # setup algorithm for instance
                if not isinstance(exp, Execution):
                    raise type_error(exp, "result of setup callable",
                                     Execution)
                # noinspection PyProtectedMember
                algo_name = sanitize_name(str(exp._algorithm))

                cd = Path(os.path.join(base_dir, algo_name, inst_name))
                cd.ensure_dir_exists()

                # generate sequence of seeds
                seeds: list[int] = [0] if warmup else \
                    rand_seeds_from_str(string=inst_name, n_seeds=runs)
                random.shuffle(seeds)
                needs_warmup = warmup or perform_warmup
                for seed in seeds:  # for every run

                    filename = sanitize_names(
                        [algo_name, inst_name, hex(seed)])
                    if warmup:
                        log_file = filename
                    else:
                        log_file = Path(
                            os.path.join(cd, filename + FILE_SUFFIX))

                        skip = True
                        if cache(log_file):
                            skip = log_file.ensure_file_exists()
                        if skip:
                            continue  # run already done

                    exp.set_rand_seed(seed)

                    if needs_warmup:  # perform warmup run
                        needs_warmup = False
                        cpy: Execution = copy.copy(exp)
                        cpy.set_max_fes(
                            pre_warmup_fes if warmup else warmup_fes, True)
                        cpy.set_max_time_millis(3600000, True)
                        cpy.set_log_file(None)
                        cpy.set_log_improvements(False)
                        cpy.set_log_all_fes(False)
                        logger(f"{wss} for {filename!r}.", thread_id)
                        with cpy.execute():
                            pass
                        del cpy

                    if warmup:
                        continue

                    exp.set_log_file(log_file)
                    logger(filename, thread_id)
                    with exp.execute() as process:  # run the experiment
                        on_completion(instance, cast(Path, log_file), process)


def __no_complete(_: Any, __: Path, ___: Process) -> None:
    """Do nothing."""


def run_experiment(
        base_dir: str, instances: Iterable[Callable[[], Any]],
        setups: Iterable[Callable[[Any], Execution]],
        n_runs: int | Iterable[int] = 11,
        perform_warmup: bool = True, warmup_fes: int = 20,
        perform_pre_warmup: bool = True, pre_warmup_fes: int = 20,
        on_completion: Callable[[Any, Path, Process], None] = __no_complete) \
        -> Path:
    """
    Run an experiment and store the log files into the given folder.

    This function will automatically run an experiment, i.e., apply a set
    `setups` of algorithm setups to a set `instances` of problem instances for
    `n_runs` each. It will collect log files and store them into an
    appropriate folder structure under the path `base_dir`. It will
    automatically draw random seeds for all algorithm runs using
    :func:`moptipy.utils.nputils.rand_seeds_from_str` based on the names of
    the problem instances to solve. This yields replicable experiments, i.e.,
    running the experiment program twice will yield exactly the same runs in
    exactly the same file structure (give and take clock-time dependent
    issues, which obviously cannot be controlled in a deterministic fashion).

    :param base_dir: the base directory where to store the results
    :param instances: an iterable of callables, each of which should return an
        object representing a problem instance, whose `__str__` representation
        is a valid name
    :param setups: an iterable of callables, each receiving an instance (as
        returned by instances) as input and producing an
        :class:`moptipy.api.execution.Execution` as output
    :param n_runs: the number of runs per algorithm-instance combination
    :param perform_warmup: should we perform a warm-up for each instance?
        If this parameter is `True`, then before the very first run of a
        thread on an instance, we will execute the algorithm for just a few
        function evaluations without logging and discard the results. The
        idea is that during this warm-up, things such as JIT compilation or
        complicated parsing can take place. While this cannot mitigate time
        measurement problems for JIT compilations taking place late in runs,
        it can at least somewhat solve the problem of delayed first FEs caused
        by compilation and parsing.
    :param warmup_fes: the number of the FEs for the warm-up runs
    :param perform_pre_warmup: should we do one warmup run for each
        instance before we begin with the actual experiments? This complements
        the warmups defined by `perform_warmup`. It could be that, for some
        reason, JIT or other activities may lead to stalls between multiple
        processes when code is encountered for the first time. This may or may
        not still cause strange timing issues even if `perform_warmup=True`.
        We therefore can do one complete round of warmups before starting the
        actual experiment. After that, we perform one garbage collection run
        and then freeze all objects surviving it to prevent them from future
        garbage collection runs. All processes that execute the experiment in
        parallel will complete their pre-warmup and only after all of them have
        completed it, the actual experiment will begin. I am not sure whether
        this makes sense or not, but it also would not hurt.
    :param pre_warmup_fes: the FEs for the pre-warmup runs
    :param on_completion: a function to be called for every completed run,
        receiving the instance, the path to the log file (before it is
        created) and the :class:`~moptipy.api.process.Process` of the run
        as parameters

    :returns: the canonicalized path to `base_dir`
    """
    if not isinstance(instances, Iterable):
        raise type_error(instances, "instances", Iterable)
    if not isinstance(setups, Iterable):
        raise type_error(setups, "setups", Iterable)
    if not isinstance(perform_warmup, bool):
        raise type_error(perform_warmup, "perform_warmup", bool)
    if not isinstance(perform_pre_warmup, bool):
        raise type_error(perform_pre_warmup, "perform_pre_warmup", bool)
    check_int_range(warmup_fes, "warmup_fes", 1, 1_000_000)
    check_int_range(pre_warmup_fes, "pre_warmup_fes", 1, 1_000_000)

    instances = list(instances)
    if list.__len__(instances) <= 0:
        raise ValueError("Instance enumeration is empty.")
    for instance in instances:
        if not callable(instance):
            raise type_error(instance, "all instances", call=True)

    if str.__len__(get_sys_info()) <= 0:
        raise ValueError("empty system info?")

    setups = list(setups)
    if list.__len__(setups) <= 0:
        raise ValueError("Setup enumeration is empty.")
    for setup in setups:
        if not callable(setup):
            raise type_error(setup, "all setups", call=True)

    experiments: Final[list[list[Callable]]] = \
        [[ii, ss] for ii in instances for ss in setups]

    del instances
    del setups

    if list.__len__(experiments) <= 0:
        raise ValueError("No experiments found?")

    n_runs = [n_runs] if isinstance(n_runs, int) else list(n_runs)
    if list.__len__(n_runs) <= 0:
        raise ValueError("No number of runs provided?")
    last = 0
    for run in n_runs:
        last = check_int_range(run, "n_runs", last + 1)

    use_dir: Final[Path] = Path(base_dir)
    use_dir.ensure_dir_exists()

    thread_id: Final[str] = f"{getpid():x}"
    logger("beginning experiment execution.", thread_id)
    __run_experiment(base_dir=use_dir,
                     experiments=experiments,
                     n_runs=n_runs,
                     thread_id=thread_id,
                     perform_warmup=perform_warmup,
                     warmup_fes=warmup_fes,
                     perform_pre_warmup=perform_pre_warmup,
                     pre_warmup_fes=pre_warmup_fes,
                     on_completion=on_completion)
    logger("finished experiment execution.", thread_id)
    return use_dir



"""
SampleStatistics aggregated over multiple instances of `EndResult`.

The :mod:`~moptipy.evaluation.end_results` records hold the final result of
a run of an optimization algorithm on a problem instance. Often, we do not
want to compare these single results directly, but instead analyze summary
statistics, such as the mean best objective value found. For this purpose,
:class:`EndStatistics` exists. It summarizes the singular results from the
runs into a record with the most important statistics.
"""
import argparse
import os.path
from dataclasses import dataclass
from math import ceil, inf, isfinite
from typing import Any, Callable, Final, Iterable, cast

from pycommons.io.console import logger
from pycommons.io.csv import (
    SCOPE_SEPARATOR,
    csv_column,
    csv_column_or_none,
    csv_read,
    csv_scope,
    csv_select_scope,
    csv_select_scope_or_none,
    csv_str_or_none,
    csv_val_or_none,
    csv_write,
)
from pycommons.io.path import Path, file_path, line_writer
from pycommons.math.sample_statistics import (
    KEY_MEAN_ARITH,
    KEY_STDDEV,
    SampleStatistics,
    from_samples,
    from_single_value,
)
from pycommons.math.sample_statistics import CsvReader as StatReader
from pycommons.math.sample_statistics import CsvWriter as StatWriter
from pycommons.math.sample_statistics import getter as stat_getter
from pycommons.strings.string_conv import (
    num_or_none_to_str,
    str_to_num,
)
from pycommons.types import (
    check_int_range,
    reiterable,
    type_error,
    type_name_of,
)

from moptipy.api.logging import (
    KEY_ALGORITHM,
    KEY_BEST_F,
    KEY_GOAL_F,
    KEY_INSTANCE,
    KEY_LAST_IMPROVEMENT_FE,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS,
    KEY_MAX_FES,
    KEY_MAX_TIME_MILLIS,
    KEY_TOTAL_FES,
    KEY_TOTAL_TIME_MILLIS,
)
from moptipy.evaluation._utils import (
    _check_max_time_millis,
)
from moptipy.evaluation.base import (
    DESC_ALGORITHM,
    DESC_ENCODING,
    DESC_INSTANCE,
    DESC_OBJECTIVE_FUNCTION,
    F_NAME_RAW,
    F_NAME_SCALED,
    KEY_ENCODING,
    KEY_N,
    KEY_OBJECTIVE_FUNCTION,
    MultiRunData,
    motipy_footer_bottom_comments,
)
from moptipy.evaluation.end_results import (
    DESC_BEST_F,
    DESC_GOAL_F,
    DESC_LAST_IMPROVEMENT_FE,
    DESC_LAST_IMPROVEMENT_TIME_MILLIS,
    DESC_MAX_FES,
    DESC_MAX_TIME_MILLIS,
    DESC_TOTAL_FES,
    DESC_TOTAL_TIME_MILLIS,
    EndResult,
)
from moptipy.evaluation.end_results import from_csv as end_results_from_csv
from moptipy.evaluation.end_results import from_logs as end_results_from_logs
from moptipy.utils.help import moptipy_argparser
from moptipy.utils.math import try_int, try_int_div

#: The key for the best F.
KEY_BEST_F_SCALED: Final[str] = KEY_BEST_F + "scaled"
#: The key for the number of successful runs.
KEY_N_SUCCESS: Final[str] = "successN"
#: The key for the success FEs.
KEY_SUCCESS_FES: Final[str] = "successFEs"
#: The key for the success time millis.
KEY_SUCCESS_TIME_MILLIS: Final[str] = "successTimeMillis"
#: The key for the ERT in FEs.
KEY_ERT_FES: Final[str] = "ertFEs"
#: The key for the ERT in milliseconds.
KEY_ERT_TIME_MILLIS: Final[str] = "ertTimeMillis"


@dataclass(frozen=True, init=False, order=False, eq=False)
class EndStatistics(MultiRunData):
    """
    Statistics over end results of one or multiple algorithm*instance setups.

    If one algorithm*instance is used, then `algorithm` and `instance` are
    defined. Otherwise, only the parameter which is the same over all recorded
    runs is defined.
    """

    #: The statistics about the best encountered result.
    best_f: SampleStatistics
    #: The statistics about the last improvement FE.
    last_improvement_fe: SampleStatistics
    #: The statistics about the last improvement time.
    last_improvement_time_millis: SampleStatistics
    #: The statistics about the total number of FEs.
    total_fes: SampleStatistics
    #: The statistics about the total time.
    total_time_millis: SampleStatistics
    #: The goal objective value.
    goal_f: SampleStatistics | int | float | None
    #: best_f / goal_f if goal_f is consistently defined and always positive.
    best_f_scaled: SampleStatistics | None
    #: The number of successful runs, if goal_f != None, else None.
    n_success: int | None
    #: The FEs to success, if n_success > 0, None otherwise.
    success_fes: SampleStatistics | None
    #: The time to success, if n_success > 0, None otherwise.
    success_time_millis: SampleStatistics | None
    #: The ERT if FEs, while is inf if n_success=0, None if goal_f is None,
    #: and finite otherwise.
    ert_fes: int | float | None
    #: The ERT if milliseconds, while is inf if n_success=0, None if goal_f
    #: is None, and finite otherwise.
    ert_time_millis: int | float | None
    #: The budget in FEs, if every run had one; None otherwise.
    max_fes: SampleStatistics | int | None
    #: The budget in milliseconds, if every run had one; None otherwise.
    max_time_millis: SampleStatistics | int | None

    def __init__(self,
                 algorithm: str | None,
                 instance: str | None,
                 objective: str | None,
                 encoding: str | None,
                 n: int,
                 best_f: SampleStatistics,
                 last_improvement_fe: SampleStatistics,
                 last_improvement_time_millis: SampleStatistics,
                 total_fes: SampleStatistics,
                 total_time_millis: SampleStatistics,
                 goal_f: float | int | SampleStatistics | None,
                 best_f_scaled: SampleStatistics | None,
                 n_success: int | None,
                 success_fes: SampleStatistics | None,
                 success_time_millis: SampleStatistics | None,
                 ert_fes: int | float | None,
                 ert_time_millis: int | float | None,
                 max_fes: SampleStatistics | int | None,
                 max_time_millis: SampleStatistics | int | None):
        """
        Create the end statistics of an experiment-setup combination.

        :param algorithm: the algorithm name, if all runs are with the same
            algorithm
        :param instance: the instance name, if all runs are on the same
            instance
        :param objective: the objective name, if all runs are on the same
            objective function, `None` otherwise
        :param encoding: the encoding name, if all runs are on the same
            encoding and an encoding was actually used, `None` otherwise
        :param n: the total number of runs
        :param best_f: statistics about the best achieved result
        :param last_improvement_fe: statistics about the last improvement FE
        :param last_improvement_time_millis: statistics about the last
            improvement time
        :param total_fes: statistics about the total FEs
        :param total_time_millis: statistics about the total runtime in
            milliseconds
        :param goal_f: if the goal objective value is not defined sometimes,
            this will be `None`. If it is always defined and always the same,
            then this will be that value. If different goal values exist, then
            this is the `SampleStatistics` record about them
        :param best_f_scaled: if `goal_f` is not `None` and greater than zero,
            then here we provide statistics about `best_f` divided by the
            corresponding `goal_f`
        :param n_success: the number of successful runs is only defined if
            `goal_f` is not `None` and counts the number of runs that reach or
            surpass their corresponding `goal_f`
        :param success_fes: if `goal_f` is not `None`,
            then this holds statistics about the last improvement FE of only
            the successful runs
        :param success_time_millis: if `goal_f` is not `None`, then this holds
            statistics about the last improvement times of only the successful
            runs
        :param ert_fes: if `goal_f` is always defined, then this is the
            empirically estimated running time to solve the problem in FEs if
            `n_success>0` and `inf` otherwise
        :param ert_time_millis: if `goal_f` is always defined, then this is
            the empirically estimated running time to solve the problem in
            milliseconds if `n_success>0` and `inf` otherwise
        :param max_fes: the budget in FEs, if any
        :param max_time_millis: the budget in terms of milliseconds
        """
        super().__init__(algorithm, instance, objective, encoding, n)

        if not isinstance(best_f, SampleStatistics):
            raise type_error(best_f, "best_f", SampleStatistics)
        object.__setattr__(self, "best_f", best_f)
        if best_f.n != n:
            raise ValueError(f"best_f.n={best_f.n} != n={n}")

        if not isinstance(last_improvement_fe, SampleStatistics):
            raise type_error(last_improvement_fe, "last_improvement_fe",
                             SampleStatistics)
        if last_improvement_fe.n != n:
            raise ValueError(
                f"last_improvement_fe.n={last_improvement_fe.n} != n={n}")
        check_int_range(
            last_improvement_fe.minimum, "last_improvement_fe.minimum",
            1, 1_000_000_000_000_000)
        check_int_range(
            last_improvement_fe.maximum, "last_improvement_fe.maximum",
            last_improvement_fe.minimum, 1_000_000_000_000_000)
        object.__setattr__(self, "last_improvement_fe", last_improvement_fe)

        if not isinstance(last_improvement_time_millis, SampleStatistics):
            raise type_error(last_improvement_time_millis,
                             "last_improvement_time_millis", SampleStatistics)
        if last_improvement_time_millis.n != n:
            raise ValueError("last_improvement_time_millis.n="
                             f"{last_improvement_time_millis.n} != n={n}")
        check_int_range(
            last_improvement_time_millis.minimum,
            "last_improvement_time_millis.minimum",
            0, 100_000_000_000)
        check_int_range(
            last_improvement_time_millis.maximum,
            "last_improvement_time_millis.maximum",
            last_improvement_time_millis.minimum, 100_000_000_000)
        object.__setattr__(self, "last_improvement_time_millis",
                           last_improvement_time_millis)

        if not isinstance(total_fes, SampleStatistics):
            raise type_error(total_fes, "total_fes", SampleStatistics)
        if total_fes.n != n:
            raise ValueError(
                f"total_fes.n={total_fes.n} != n={n}")
        check_int_range(
            total_fes.minimum, "total_fes.minimum",
            last_improvement_fe.minimum, 1_000_000_000_000_000)
        check_int_range(
            total_fes.maximum, "total_fes.maximum",
            max(total_fes.minimum, last_improvement_fe.maximum),
            1_000_000_000_000_000)
        object.__setattr__(self, "total_fes", total_fes)

        if not isinstance(total_time_millis, SampleStatistics):
            raise type_error(total_time_millis, "total_time_millis",
                             SampleStatistics)
        if total_time_millis.n != n:
            raise ValueError(
                f"total_time_millis.n={total_time_millis.n} != n={n}")
        check_int_range(
            total_time_millis.minimum, "total_time_millis.minimum",
            last_improvement_time_millis.minimum, 100_000_000_000)
        check_int_range(
            total_time_millis.maximum, "total_time_millis.maximum",
            max(total_time_millis.minimum,
                last_improvement_time_millis.maximum),
            100_000_000_000)
        object.__setattr__(self, "total_time_millis", total_time_millis)

        if goal_f is None:
            if best_f_scaled is not None:
                raise ValueError(
                    "If goal_f is None, best_f_scaled must also be None, "
                    f"but is {type(best_f_scaled)}.")
            if n_success is not None:
                raise ValueError(
                    "If goal_f is None, n_success must also be None, "
                    f"but is {type(n_success)}.")
            if success_fes is not None:
                raise ValueError(
                    "If success_fes is None, best_f_scaled must also be None, "
                    f"but is {type(success_fes)}.")
            if success_time_millis is not None:
                raise ValueError(
                    "If success_time_millis is None, best_f_scaled "
                    "must also be None, "
                    f"but is {type(success_time_millis)}.")
            if ert_fes is not None:
                raise ValueError(
                    "If goal_f is None, ert_fes must also be None, "
                    f"but is {type(ert_fes)}.")
            if ert_time_millis is not None:
                raise ValueError(
                    "If goal_f is None, ert_time_millis must also be None, "
                    f"but is {type(ert_time_millis)}.")
        else:  # goal_f is not None
            if isinstance(goal_f, SampleStatistics):
                if goal_f.n != n:
                    raise ValueError(f"goal_f.n={goal_f.n} != n={n}")
                goal_f = goal_f.compact(False)
            if isinstance(goal_f, float):
                goal_f = None if goal_f <= (-inf) else try_int(goal_f)
            elif not isinstance(goal_f, int | SampleStatistics):
                raise type_error(goal_f, "goal_f", (
                    int, float, SampleStatistics))

            if best_f_scaled is not None:
                goal_f_min: Final[int | float] = \
                    goal_f.minimum if isinstance(goal_f, SampleStatistics) \
                    else goal_f
                if goal_f_min <= 0:
                    raise ValueError(
                        f"best_f_scaled must be None if minimum goal_f "
                        f"({goal_f_min}) of goal_f {goal_f} is not positive,"
                        f" but is {best_f_scaled}.")
                if not isinstance(best_f_scaled, SampleStatistics):
                    raise type_error(best_f_scaled, "best_f_scaled",
                                     SampleStatistics)
                if best_f_scaled.n != n:
                    raise ValueError(
                        f"best_f_scaled.n={best_f_scaled.n} != n={n}")
                if best_f_scaled.minimum < 0:
                    raise ValueError(
                        "best_f_scaled cannot be negative, but encountered "
                        f"{best_f_scaled.minimum}.")

            check_int_range(n_success, "n_success")
            if not isinstance(ert_fes, int | float):
                raise type_error(ert_fes, "ert_fes", (int, float))
            if not isinstance(ert_time_millis, int | float):
                raise type_error(ert_time_millis, "ert_time_millis",
                                 (int, float))

            if n_success > 0:
                if not isinstance(success_fes, SampleStatistics):
                    raise type_error(success_fes,
                                     "if n_success>0, then success_fes",
                                     SampleStatistics)
                if success_fes.n != n_success:
                    raise ValueError(f"success_fes.n={success_fes.n} != "
                                     f"n_success={n_success}")
                check_int_range(
                    success_fes.minimum, "success_fes.minimum",
                    last_improvement_fe.minimum, 1_000_000_000_000_000)
                check_int_range(
                    success_fes.maximum, "success_fes.maximum",
                    success_fes.minimum, last_improvement_fe.maximum)
                if not isinstance(success_time_millis, SampleStatistics):
                    raise type_error(
                        success_time_millis,
                        "if n_success>0, then success_time_millis",
                        SampleStatistics)
                if success_time_millis.n != n_success:
                    raise ValueError(
                        f"success_time_millis.n={success_time_millis.n} != "
                        f"n_success={n_success}")
                check_int_range(
                    success_time_millis.minimum,
                    "success_time_millis.minimum",
                    last_improvement_time_millis.minimum, 100_000_000_000)
                check_int_range(
                    success_time_millis.maximum,
                    "success_time_millis.maximum",
                    success_time_millis.minimum,
                    last_improvement_time_millis.maximum)
                ert_fes = try_int(ert_fes)
                if ert_fes < success_fes.minimum:
                    raise ValueError(
                        "ert_fes must be >= "
                        f"{success_fes.minimum}, but is {ert_fes}.")
                ert_fe_max = ceil(total_fes.mean_arith * n)
                if ert_fes > ert_fe_max:
                    raise ValueError(
                        "ert_fes must be <= "
                        f"{ert_fe_max}, but is {ert_fes}.")

                ert_time_millis = try_int(ert_time_millis)
                if ert_time_millis < success_time_millis.minimum:
                    raise ValueError(
                        "ert_time_millis must be >= "
                        f"{success_time_millis.minimum}, but "
                        f"is {ert_time_millis}.")
                ert_time_max = ceil(total_time_millis.mean_arith * n)
                if ert_time_millis > ert_time_max:
                    raise ValueError(
                        "ert_time_millis must be <= "
                        f"{ert_time_max}, but is {ert_time_millis}.")
            else:
                if success_fes is not None:
                    raise ValueError(
                        "If n_success<=0, then success_fes must be None, "
                        f"but it's a {type_name_of(success_fes)}.")
                if success_time_millis is not None:
                    raise ValueError(
                        "If n_success<=0, then success_time_millis must be "
                        f"None, but it is a "
                        f"{type_name_of(success_time_millis)}.")
                if ert_fes < inf:
                    raise ValueError(
                        "If n_success<=0, then ert_fes must "
                        f"be inf, but it's {ert_fes}.")
                if ert_time_millis < inf:
                    raise ValueError(
                        "If n_success<=0, then ert_time_millis must "
                        f"be inf, but it's {ert_time_millis}.")

        object.__setattr__(self, "goal_f", goal_f)
        object.__setattr__(self, "best_f_scaled", best_f_scaled)
        object.__setattr__(self, "n_success", n_success)
        object.__setattr__(self, "success_fes", success_fes)
        object.__setattr__(self, "success_time_millis", success_time_millis)
        object.__setattr__(self, "ert_fes", ert_fes)
        object.__setattr__(self, "ert_time_millis", ert_time_millis)

        if isinstance(max_fes, SampleStatistics):
            if max_fes.n != n:
                raise ValueError(f"max_fes.n={max_fes.n} != n={n}")
            max_fes_f: int | float | SampleStatistics = max_fes.compact(
                needs_n=False)
            if isinstance(max_fes_f, float):
                raise type_error(max_fes_f, "max_fes", (
                    int, SampleStatistics, None))
            max_fes = max_fes_f
        if isinstance(max_fes, int):
            if (max_fes < total_fes.maximum) or (max_fes < 0):
                raise ValueError(f"0<max_fes must be >= "
                                 f"{total_fes.maximum}, but is {max_fes}.")
        elif isinstance(max_fes, SampleStatistics):
            if (max_fes.minimum < total_fes.minimum) or (
                    max_fes.minimum <= 0):
                raise ValueError(
                    f"0<max_fes.minimum must be >= {total_fes.minimum},"
                    f" but is {max_fes.minimum}.")
            if max_fes.maximum < total_fes.maximum:
                raise ValueError(
                    f"max_fes.maximum must be >= {total_fes.maximum},"
                    f" but is {max_fes.maximum}.")
        elif max_fes is not None:
            raise type_error(max_fes, "max_fes", (int, SampleStatistics, None))
        object.__setattr__(self, "max_fes", max_fes)

        if isinstance(max_time_millis, SampleStatistics):
            if max_time_millis.n != n:
                raise ValueError(
                    f"max_time_millis.n={max_time_millis.n} != n={n}")
            max_time_millis_f: int | float | SampleStatistics = (
                max_time_millis.compact(False))
            if isinstance(max_time_millis_f, float):
                raise type_error(max_time_millis_f, "max_time_millis", (
                    int, SampleStatistics, None))
        if isinstance(max_time_millis, int):
            _check_max_time_millis(max_time_millis,
                                   total_fes.minimum,
                                   total_time_millis.maximum)
        elif isinstance(max_time_millis, SampleStatistics):
            _check_max_time_millis(max_time_millis.minimum,
                                   total_fes.minimum,
                                   total_time_millis.minimum)
            _check_max_time_millis(max_time_millis.maximum,
                                   total_fes.minimum,
                                   total_time_millis.maximum)
        elif max_time_millis is not None:
            raise type_error(max_time_millis, "max_time_millis",
                             (int, SampleStatistics, None))
        object.__setattr__(self, "max_time_millis", max_time_millis)

    def get_n(self) -> int:
        """
        Get the number of runs.

        :returns: the number of runs.
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.n

    def get_best_f(self) -> SampleStatistics:
        """
        Get the statistics about the best objective value reached.

        :returns: the statistics about the best objective value reached
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.best_f

    def get_last_improvement_fe(self) -> SampleStatistics:
        """
        Get the statistics about the last improvement FE.

        :returns: the statistics about the last improvement FE
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.last_improvement_fe

    def get_last_improvement_time_millis(self) -> SampleStatistics:
        """
        Get the statistics about the last improvement time millis.

        :returns: the statistics about the last improvement time millis
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.last_improvement_time_millis

    def get_total_fes(self) -> SampleStatistics:
        """
        Get the statistics about the total FEs.

        :returns: the statistics about the total FEs
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.total_fes

    def get_total_time_millis(self) -> SampleStatistics:
        """
        Get the statistics about the total time millis.

        :returns: the statistics about the total time millis
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.total_time_millis

    def get_goal_f(self) -> SampleStatistics | int | float | None:
        """
        Get the statistics about the goal objective value.

        :returns: the statistics about the goal objective value
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.goal_f

    def get_best_f_scaled(self) -> SampleStatistics | None:
        """
        Get the statistics about the scaled best objective value.

        :returns: the statistics about the scaled best objective value
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.best_f_scaled

    def get_n_success(self) -> int | None:
        """
        Get the number of successful runs.

        :returns: the number of successful runs.
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.n_success

    def get_success_fes(self) -> SampleStatistics | None:
        """
        Get the statistics about the FEs until success of the successful runs.

        :returns: the statistics about the FEs until success of the successful
            runs
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.success_fes

    def get_success_time_millis(self) -> SampleStatistics | None:
        """
        Get the statistics about the ms until success of the successful runs.

        :returns: the statistics about the ms until success of the successful
            runs
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.success_time_millis

    def get_ert_fes(self) -> int | float | None:
        """
        Get the expected FEs until success.

        :returns: the statistics about the expected FEs until success.
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.ert_fes

    def get_ert_time_millis(self) -> int | float | None:
        """
        Get the expected milliseconds until success.

        :returns: the statistics about the expected milliseconds until
            success.
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.ert_time_millis

    def get_max_fes(self) -> SampleStatistics | int | None:
        """
        Get the statistics about the maximum permitted FEs.

        :returns: the statistics about the maximum permitted FEs
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.max_fes

    def get_max_time_millis(self) -> SampleStatistics | int | None:
        """
        Get the statistics about the maximum permitted runtime in ms.

        :returns: the statistics about the maximum permitted runtime in ms
        """
        if not isinstance(self, EndStatistics):
            raise type_error(self, "self", EndStatistics)
        return self.max_time_millis


def create(source: Iterable[EndResult]) -> EndStatistics:
    """
    Create an `EndStatistics` Record from an Iterable of `EndResult`.

    :param source: the source
    :return: the statistics
    :rtype: EndStatistics
    """
    if not isinstance(source, Iterable):
        raise type_error(source, "source", Iterable)

    n: int = 0
    best_f: list[int | float] = []
    last_improvement_fe: list[int] = []
    last_improvement_time_millis: list[int] = []
    total_fes: list[int] = []
    total_time_millis: list[int] = []
    max_fes: list[int] | None = []
    max_fes_same: bool = True
    max_time_millis: list[int] | None = []
    max_time_same: bool = True
    goal_f: list[int | float] | None = []
    goal_f_same: bool = True
    best_f_scaled: list[float] | None = []
    n_success: int | None = 0
    success_fes: list[int] | None = []
    success_times: list[int] | None = []

    fes: int = 0
    time: int = 0
    algorithm: str | None = None
    instance: str | None = None
    objective: str | None = None
    encoding: str | None = None

    for er in source:
        if not isinstance(er, EndResult):
            raise type_error(er, "end result", EndResult)
        if n == 0:
            algorithm = er.algorithm
            instance = er.instance
            objective = er.objective
            encoding = er.encoding
        else:
            if algorithm != er.algorithm:
                algorithm = None
            if instance != er.instance:
                instance = None
            if objective != er.objective:
                objective = None
            if encoding != er.encoding:
                encoding = None
        n += 1
        best_f.append(er.best_f)
        last_improvement_fe.append(er.last_improvement_fe)
        last_improvement_time_millis.append(
            er.last_improvement_time_millis)
        total_fes.append(er.total_fes)
        total_time_millis.append(er.total_time_millis)
        if er.max_fes is None:
            max_fes = None
        elif max_fes is not None:
            if n > 1:
                max_fes_same = max_fes_same \
                    and (max_fes[-1] == er.max_fes)
            max_fes.append(er.max_fes)
        if er.max_time_millis is None:
            max_time_millis = None
        elif max_time_millis is not None:
            if n > 1:
                max_time_same = \
                    max_time_same \
                    and (max_time_millis[-1] == er.max_time_millis)
            max_time_millis.append(er.max_time_millis)

        if er.goal_f is None:
            goal_f = None
            best_f_scaled = None
            n_success = None
            success_fes = None
            success_times = None
        elif goal_f is not None:
            if n > 1:
                goal_f_same = goal_f_same and (goal_f[-1] == er.goal_f)
            goal_f.append(er.goal_f)

            if er.goal_f <= 0:
                best_f_scaled = None
            elif best_f_scaled is not None:
                best_f_scaled.append(er.best_f / er.goal_f)

            if er.best_f <= er.goal_f:
                n_success += 1
                success_fes.append(er.last_improvement_fe)
                success_times.append(er.last_improvement_time_millis)
                fes += er.last_improvement_fe
                time += er.last_improvement_time_millis
            else:
                fes += er.total_fes
                time += er.total_time_millis
    if n <= 0:
        raise ValueError("There must be at least one end result record.")

    return EndStatistics(
        algorithm,
        instance,
        objective,
        encoding,
        n,
        from_samples(best_f),
        from_samples(last_improvement_fe),
        from_samples(last_improvement_time_millis),
        from_samples(total_fes),
        from_samples(total_time_millis),
        None if (goal_f is None)
        else (goal_f[0] if goal_f_same else from_samples(goal_f)),
        None if (best_f_scaled is None)
        else from_samples(best_f_scaled),
        n_success,
        None if (n_success is None) or (n_success <= 0)
        else from_samples(success_fes),
        None if (n_success is None) or (n_success <= 0)
        else from_samples(success_times),
        None if (n_success is None)
        else (inf if (n_success <= 0) else try_int_div(fes, n_success)),
        None if (n_success is None) else
        (inf if (n_success <= 0) else try_int_div(time, n_success)),
        None if max_fes is None else
        (max_fes[0] if max_fes_same else from_samples(max_fes)),
        None if max_time_millis is None
        else (max_time_millis[0] if max_time_same
              else from_samples(max_time_millis)))


def from_end_results(source: Iterable[EndResult],
                     consumer: Callable[[EndStatistics], Any],
                     join_all_algorithms: bool = False,
                     join_all_instances: bool = False,
                     join_all_objectives: bool = False,
                     join_all_encodings: bool = False) -> None:
    """
    Aggregate statistics over a stream of end results.

    :param source: the stream of end results
    :param consumer: the destination to which the new records will be
        sent, can be the `append` method of a :class:`list`
    :param join_all_algorithms: should the statistics be aggregated
        over all algorithms
    :param join_all_instances: should the statistics be aggregated
        over all algorithms
    :param join_all_objectives: should the statistics be aggregated over
        all objectives?
    :param join_all_encodings: should statistics be aggregated over all
        encodings
    """
    if not isinstance(source, Iterable):
        raise type_error(source, "source", Iterable)
    if not callable(consumer):
        raise type_error(consumer, "consumer", call=True)
    if not isinstance(join_all_algorithms, bool):
        raise type_error(join_all_algorithms,
                         "join_all_algorithms", bool)
    if not isinstance(join_all_instances, bool):
        raise type_error(join_all_instances, "join_all_instances", bool)
    if not isinstance(join_all_objectives, bool):
        raise type_error(join_all_objectives, "join_all_objectives", bool)
    if not isinstance(join_all_encodings, bool):
        raise type_error(join_all_encodings, "join_all_encodings", bool)

    if (join_all_algorithms and join_all_instances
            and join_all_objectives and join_all_encodings):
        consumer(create(source))
        return

    sorter: dict[tuple[str, str, str, str], list[EndResult]] = {}
    for er in source:
        if not isinstance(er, EndResult):
            raise type_error(source, "end results from source",
                             EndResult)
        key = ("" if join_all_algorithms else er.algorithm,
               "" if join_all_instances else er.instance,
               "" if join_all_objectives else er.objective,
               "" if join_all_encodings else (
                   "" if er.encoding is None else er.encoding))
        if key in sorter:
            lst = sorter[key]
        else:
            lst = []
            sorter[key] = lst
        lst.append(er)

    if len(sorter) <= 0:
        raise ValueError("source must not be empty")

    if len(sorter) > 1:
        for key in sorted(sorter.keys()):
            consumer(create(sorter[key]))
    else:
        consumer(create(next(iter(sorter.values()))))


def to_csv(data: EndStatistics | Iterable[EndStatistics],
           file: str) -> Path:
    """
    Store a set of :class:`EndStatistics` in a CSV file.

    :param data: the data to store
    :param file: the file to generate
    :return: the path to the generated CSV file
    """
    path: Final[Path] = Path(file)
    logger(f"Writing end result statistics to CSV file {path!r}.")
    path.ensure_parent_dir_exists()
    with path.open_for_write() as wt:
        csv_write(
            data=[data] if isinstance(data, EndStatistics) else sorted(data),
            consumer=line_writer(wt),
            setup=CsvWriter().setup,
            get_column_titles=CsvWriter.get_column_titles,
            get_row=CsvWriter.get_row,
            get_header_comments=CsvWriter.get_header_comments,
            get_footer_comments=CsvWriter.get_footer_comments,
            get_footer_bottom_comments=CsvWriter.get_footer_bottom_comments)
    logger(f"Done writing end result statistics to CSV file {path!r}.")
    return path


def from_csv(file: str,
             consumer: Callable[[EndStatistics], Any]) -> None:
    """
    Parse a CSV file and collect all encountered :class:`EndStatistics`.

    :param file: the file to parse
    :param consumer: the consumer to receive all the parsed instances of
        :class:`~moptipy.evaluation.end_statistics.EndStatistics`, can be
        the `append` method of a :class:`list`
    """
    path: Final[Path] = file_path(file)
    logger(f"Begin reading end result statistics from CSV file {path!r}.")
    with path.open_for_read() as rd:
        csv_read(rows=rd,
                 setup=CsvReader,
                 parse_row=CsvReader.parse_row,
                 consumer=consumer)
    logger("Finished reading end result statistics from CSV "
           f"file {path!r}.")


#: the internal getters that can work directly
__PROPERTIES: Final[Callable[[str], Callable[[
    EndStatistics], SampleStatistics | int | float | None] | None]] = {
    KEY_N: EndStatistics.get_n,
    KEY_N_SUCCESS: EndStatistics.get_n_success,
    KEY_ERT_FES: EndStatistics.get_ert_fes,
    KEY_ERT_TIME_MILLIS: EndStatistics.get_ert_time_millis,
    KEY_GOAL_F: EndStatistics.get_goal_f,
    KEY_MAX_TIME_MILLIS: EndStatistics.get_max_time_millis,
    KEY_MAX_FES: EndStatistics.get_max_fes,
    KEY_BEST_F: EndStatistics.get_best_f,
    F_NAME_RAW: EndStatistics.get_best_f,
    KEY_LAST_IMPROVEMENT_FE: EndStatistics.get_last_improvement_fe,
    "last improvement FE": EndStatistics.get_last_improvement_fe,
    KEY_LAST_IMPROVEMENT_TIME_MILLIS:
        EndStatistics.get_last_improvement_time_millis,
    "last improvement ms": EndStatistics.get_last_improvement_time_millis,
    KEY_BEST_F_SCALED: EndStatistics.get_best_f_scaled,
    KEY_SUCCESS_FES: EndStatistics.get_success_fes,
    KEY_SUCCESS_TIME_MILLIS: EndStatistics.get_success_time_millis,
    F_NAME_SCALED: EndStatistics.get_best_f_scaled,
    KEY_TOTAL_FES: EndStatistics.get_total_fes,
    "fes": EndStatistics.get_total_fes,
    KEY_TOTAL_TIME_MILLIS: EndStatistics.get_total_time_millis,
    "ms": EndStatistics.get_total_time_millis,
    "f": EndStatistics.get_best_f,
    "budgetFEs": EndStatistics.get_max_fes,
    "budgetMS": EndStatistics.get_max_time_millis,
}.get

#: the success keys
__SUCCESS_KEYS: Final[Callable[[str], bool]] = {
    KEY_SUCCESS_FES, KEY_SUCCESS_TIME_MILLIS,
}.__contains__

#: the internal static getters
__STATIC: Final[dict[str, Callable[[EndStatistics], int | float | None]]] = {
    KEY_N: EndStatistics.get_n,
    KEY_N_SUCCESS: EndStatistics.get_n_success,
    KEY_ERT_FES: EndStatistics.get_ert_fes,
    KEY_ERT_TIME_MILLIS: EndStatistics.get_ert_time_millis,
}


def getter(dimension: str) -> Callable[[EndStatistics], int | float | None]:
    """
    Create a function that obtains the given dimension from EndStatistics.

    :param dimension: the dimension
    :returns: a callable that returns the value corresponding to the
        dimension
    """
    dimension = str.strip(dimension)
    direct: Callable[[EndStatistics], int | float | None] = \
        __STATIC.get(dimension)
    if direct is not None:
        return direct

    names: Final[list[str]] = str.split(str.strip(dimension), SCOPE_SEPARATOR)
    n_names: Final[int] = list.__len__(names)
    if not (0 < n_names < 3):
        raise ValueError(
            f"Invalid name combination {dimension!r} -> {names!r}.")
    getter_1: Final[Callable[[
        EndStatistics], int | float | SampleStatistics | None] | None] = \
        __PROPERTIES(names[0])
    if getter_1 is None:
        raise ValueError(f"Invalid dimension {names[0]!r} in {dimension!r}.")
    getter_2: Final[Callable[[
        SampleStatistics], int | float | None]] = \
        stat_getter(names[1] if n_names > 1 else KEY_MEAN_ARITH)

    if getter_2 is stat_getter(KEY_STDDEV):  # it is sd
        n_prop: Final[Callable[[EndStatistics], int | None]] = \
            EndStatistics.get_n_success if __SUCCESS_KEYS(
                names[0]) else EndStatistics.get_n

        def __combo_sd(
                data: EndStatistics, __g1=getter_1, __g2=getter_2,
                __n=n_prop) -> int | float | None:
            val: int | float | SampleStatistics | None = __g1(data)
            if val is None:
                return None
            if isinstance(val, int | float):
                n = __n(data)
                return None if (n is None) or (n <= 0) else 0
            return __g2(val)
        direct = cast(Callable[[EndStatistics], int | float | None],
                      __combo_sd)
    else:  # any other form of mean or statistic

        def __combo_no_sd(data: EndStatistics,
                          __g1=getter_1, __g2=getter_2) -> int | float | None:
            val: int | float | SampleStatistics | None = __g1(data)
            if (val is None) or (isinstance(val, int | float)):
                return val
            return __g2(val)
        direct = cast(Callable[[EndStatistics], int | float | None],
                      __combo_no_sd)

    __STATIC[dimension] = direct
    return direct


def _to_csv_writer(
        get_func: Callable[
            [EndStatistics], SampleStatistics | int | float | None],
        n_func: Callable[[EndStatistics], int],
        data: Iterable[EndStatistics],
        scope: str | None = None,
        what_short: str | None = None,
        what_long: str | None = None) -> StatWriter | None:
    """
    Get a CSV Writer for the given data subset.

    :param get_func: the getter for the value
    :param n_func: the n-getter
    :param data: the data iterator
    :param scope: the scope to use
    :param what_short: the short description
    :param what_long: the long description
    :returns: the writer, if there was any associated data
    """
    refined: list[tuple[SampleStatistics | int | float | None, int]] = [
        v for v in ((get_func(es), n_func(es)) for es in data)
        if v[0] is not None]
    if list.__len__(refined) <= 0:
        return None
    return StatWriter(scope=scope, n_not_needed=True, what_short=what_short,
                      what_long=what_long).setup((
                          from_single_value(v, n) for v, n in refined))


class CsvWriter:
    """A class for CSV writing of :class:`EndStatistics`."""

    def __init__(self, scope: str | None = None) -> None:
        """
        Initialize the csv writer.

        :param scope: the prefix to be pre-pended to all columns
        """
        #: an optional scope
        self.scope: Final[str | None] = (
            str.strip(scope)) if scope is not None else None

        #: has this writer been set up?
        self.__setup: bool = False
        #: do we put the algorithm column?
        self.__has_algorithm: bool = False
        #: do we put the instance column?
        self.__has_instance: bool = False
        #: do we put the objective column?
        self.__has_objective: bool = False
        #: do we put the encoding column?
        self.__has_encoding: bool = False
        #: do we put the goal_f column?
        self.__goal_f: StatWriter | None = None
        #: the best objective value reached
        self.__best_f: Final[StatWriter] = StatWriter(
            csv_scope(scope, KEY_BEST_F), True, KEY_BEST_F,
            "the best objective value reached per run")
        #: the FE when the last improvement happened
        self.__life: Final[StatWriter] = StatWriter(
            csv_scope(scope, KEY_LAST_IMPROVEMENT_FE), True,
            KEY_LAST_IMPROVEMENT_FE,
            "the FE when the last improvement happened in a run",
        )
        #: the milliseconds when the last improvement happened
        self.__lims: Final[StatWriter] = StatWriter(
            csv_scope(scope, KEY_LAST_IMPROVEMENT_TIME_MILLIS), True,
            KEY_LAST_IMPROVEMENT_TIME_MILLIS,
            "the millisecond when the last improvement happened in a run",
        )
        #: the total FEs
        self.__total_fes: Final[StatWriter] = StatWriter(
            csv_scope(scope, KEY_TOTAL_FES), True,
            KEY_TOTAL_FES,
            "the total FEs consumed by the runs",
        )
        #: the total milliseconds
        self.__total_ms: Final[StatWriter] = StatWriter(
            csv_scope(scope, KEY_TOTAL_TIME_MILLIS), True,
            KEY_TOTAL_TIME_MILLIS,
            "the total millisecond consumed by a run",
        )
        #: do we put the best-f-scaled column?
        self.__best_f_scaled: StatWriter | None = None
        #: do we put the n_success column?
        self.__has_n_success: bool = False
        #: do we put the success fes column?
        self.__success_fes: StatWriter | None = None
        #: do we put the success time millis column?
        self.__success_time_millis: StatWriter | None = None
        #: do we put the ert-fes column?
        self.__has_ert_fes: bool = False
        #: do we put the ert time millis column?
        self.__has_ert_time_millis: bool = False
        #: do we put the max-fes column?
        self.__max_fes: StatWriter | None = None
        #: do we put the max time millis column?
        self.__max_time_millis: StatWriter | None = None

    def setup(self, data: Iterable[EndStatistics]) -> "CsvWriter":
        """
        Set up this csv writer based on existing data.

        :param data: the data to setup with
        :returns: this writer
        """
        if self.__setup:
            raise ValueError(
                "EndStatistics CsvWriter has already been set up.")
        self.__setup = True

        data = reiterable(data)

        checker: int = 127
        for es in data:
            if es.algorithm is not None:
                self.__has_algorithm = True
                checker &= ~1
            if es.instance is not None:
                self.__has_instance = True
                checker &= ~2
            if es.objective is not None:
                self.__has_objective = True
                checker &= ~4
            if es.encoding is not None:
                self.__has_encoding = True
                checker &= ~8
            if es.n_success is not None:
                self.__has_n_success = True
                checker &= ~16
            if es.ert_fes is not None:
                self.__has_ert_fes = True
                checker &= ~32
            if es.ert_time_millis is not None:
                self.__has_ert_time_millis = True
                checker &= ~64
            if checker == 0:
                break

        scope: Final[str | None] = self.scope
        self.__goal_f = _to_csv_writer(
            EndStatistics.get_goal_f, EndStatistics.get_n,
            data, csv_scope(scope, KEY_GOAL_F),
            KEY_GOAL_F,
            "the goal objective value after which the runs can stop")
        self.__best_f_scaled = _to_csv_writer(
            EndStatistics.get_best_f_scaled, EndStatistics.get_n,
            data, csv_scope(scope, KEY_BEST_F_SCALED), KEY_BEST_F_SCALED,
            f"best objective value reached ({KEY_BEST_F}), divided by"
            f" the goal objective value ({KEY_GOAL_F})")
        self.__success_fes = _to_csv_writer(
            EndStatistics.get_success_fes, EndStatistics.get_n_success,
            data, csv_scope(scope, KEY_SUCCESS_FES), KEY_SUCCESS_FES,
            f"the FEs needed to reach {KEY_GOAL_F} for the successful runs")
        self.__success_time_millis = _to_csv_writer(
            EndStatistics.get_success_time_millis,
            EndStatistics.get_n_success, data, csv_scope(
                scope, KEY_SUCCESS_TIME_MILLIS), KEY_SUCCESS_TIME_MILLIS,
            f"the milliseconds needed to reach {KEY_GOAL_F} for the "
            "successful runs")
        self.__max_fes = _to_csv_writer(
            EndStatistics.get_max_fes, EndStatistics.get_n,
            data, csv_scope(scope, KEY_MAX_FES), KEY_MAX_FES,
            "the maximum number of FEs in the computational budget")
        self.__max_time_millis = _to_csv_writer(
            EndStatistics.get_max_time_millis, EndStatistics.get_n,
            data, csv_scope(scope, KEY_MAX_TIME_MILLIS), KEY_MAX_TIME_MILLIS,
            "the maximum milliseconds per run in the computational budget")
        self.__best_f.setup(map(EndStatistics.get_best_f, data))
        self.__life.setup(map(EndStatistics.get_last_improvement_fe, data))
        self.__lims.setup(map(
            EndStatistics.get_last_improvement_time_millis, data))
        self.__total_fes.setup(map(EndStatistics.get_total_fes, data))
        self.__total_ms.setup(map(EndStatistics.get_total_time_millis, data))

        return self

    def get_column_titles(self, dest: Callable[[str], None]) -> None:
        """
        Get the column titles.

        :param dest: the destination string consumer
        """
        p: Final[str] = self.scope
        if self.__has_algorithm:
            dest(csv_scope(p, KEY_ALGORITHM))
        if self.__has_instance:
            dest(csv_scope(p, KEY_INSTANCE))
        if self.__has_objective:
            dest(csv_scope(p, KEY_OBJECTIVE_FUNCTION))
        if self.__has_encoding:
            dest(csv_scope(p, KEY_ENCODING))
        dest(csv_scope(p, KEY_N))
        self.__best_f.get_column_titles(dest)
        self.__life.get_column_titles(dest)
        self.__lims.get_column_titles(dest)
        self.__total_fes.get_column_titles(dest)
        self.__total_ms.get_column_titles(dest)
        if self.__goal_f is not None:
            self.__goal_f.get_column_titles(dest)
        if self.__best_f_scaled is not None:
            self.__best_f_scaled.get_column_titles(dest)
        if self.__has_n_success:
            dest(csv_scope(p, KEY_N_SUCCESS))
        if self.__success_fes is not None:
            self.__success_fes.get_column_titles(dest)
        if self.__success_time_millis is not None:
            self.__success_time_millis.get_column_titles(dest)
        if self.__has_ert_fes:
            dest(csv_scope(p, KEY_ERT_FES))
        if self.__has_ert_time_millis:
            dest(csv_scope(p, KEY_ERT_TIME_MILLIS))
        if self.__max_fes is not None:
            self.__max_fes.get_column_titles(dest)
        if self.__max_time_millis is not None:
            self.__max_time_millis.get_column_titles(dest)

    def get_row(self, data: EndStatistics,
                dest: Callable[[str], None]) -> None:
        """
        Render a single end result record to a CSV row.

        :param data: the end result record
        :param dest: the string consumer
        """
        if self.__has_algorithm:
            dest("" if data.algorithm is None else data.algorithm)
        if self.__has_instance:
            dest("" if data.instance is None else data.instance)
        if self.__has_objective:
            dest("" if data.objective is None else data.objective)
        if self.__has_encoding:
            dest("" if data.encoding is None else data.encoding)
        dest(str(data.n))
        self.__best_f.get_row(data.best_f, dest)
        self.__life.get_row(data.last_improvement_fe, dest)
        self.__lims.get_row(data.last_improvement_time_millis, dest)
        self.__total_fes.get_row(data.total_fes, dest)
        self.__total_ms.get_row(data.total_time_millis, dest)
        if self.__goal_f is not None:
            self.__goal_f.get_optional_row(data.goal_f, dest, data.n)
        if self.__best_f_scaled is not None:
            self.__best_f_scaled.get_optional_row(
                data.best_f_scaled, dest, data.n)
        if self.__has_n_success:
            dest(str(data.n_success))
        if self.__success_fes is not None:
            self.__success_fes.get_optional_row(
                data.success_fes, dest, data.n_success)
        if self.__success_time_millis is not None:
            self.__success_time_millis.get_optional_row(
                data.success_time_millis, dest, data.n_success)
        if self.__has_ert_fes:
            dest(num_or_none_to_str(data.ert_fes))
        if self.__has_ert_time_millis:
            dest(num_or_none_to_str(data.ert_time_millis))
        if self.__max_fes is not None:
            self.__max_fes.get_optional_row(data.max_fes, dest, data.n)
        if self.__max_time_millis is not None:
            self.__max_time_millis.get_optional_row(
                data.max_time_millis, dest, data.n)

    def get_header_comments(self, dest: Callable[[str], None]) -> None:
        """
        Get any possible header comments.

        :param dest: the destination
        """
        dest("Experiment End Results Statistics")
        dest("See the description at the bottom of the file.")

    def get_footer_comments(self, dest: Callable[[str], None]) -> None:
        """
        Get any possible footer comments.

        :param dest: the destination
        """
        dest("")
        scope: Final[str | None] = self.scope

        dest("This file presents statistics gathered over multiple runs "
             "of optimization algorithms applied to problem instances.")
        if scope:
            dest("All end result statistics records start with prefix "
                 f"{scope}{SCOPE_SEPARATOR}.")
        if self.__has_algorithm:
            dest(f"{csv_scope(scope, KEY_ALGORITHM)}: {DESC_ALGORITHM}")
        if self.__has_instance:
            dest(f"{csv_scope(scope, KEY_INSTANCE)}: {DESC_INSTANCE}")
        if self.__has_objective:
            dest(f"{csv_scope(scope, KEY_OBJECTIVE_FUNCTION)}:"
                 f" {DESC_OBJECTIVE_FUNCTION}")
        if self.__has_encoding:
            dest(f"{csv_scope(scope, KEY_ENCODING)}: {DESC_ENCODING}")
        dest(f"{csv_scope(scope, KEY_N)}: the number of runs that were "
             f"performed for the given setup.")

        self.__best_f.get_footer_comments(dest)
        dest(f"In summary {csv_scope(scope, KEY_BEST_F)} is {DESC_BEST_F}.")

        self.__life.get_footer_comments(dest)
        dest(f"In summary {csv_scope(scope, KEY_LAST_IMPROVEMENT_FE)} "
             f"is {DESC_LAST_IMPROVEMENT_FE}.")

        self.__lims.get_footer_comments(dest)
        dest("In summary "
             f"{csv_scope(scope, KEY_LAST_IMPROVEMENT_TIME_MILLIS)} "
             f"is {DESC_LAST_IMPROVEMENT_TIME_MILLIS}.")

        self.__total_fes.get_footer_comments(dest)
        dest(f"In summary {csv_scope(scope, KEY_TOTAL_FES)} "
             f"is {DESC_TOTAL_FES}.")

        self.__total_ms.get_footer_comments(dest)
        dest(f"In summary {csv_scope(scope, KEY_TOTAL_TIME_MILLIS)} "
             f"is {DESC_TOTAL_TIME_MILLIS}.")

        if self.__goal_f is not None:
            self.__goal_f.get_footer_comments(dest)
            dest(f"In summary {csv_scope(scope, KEY_GOAL_F)} is"
                 f" {DESC_GOAL_F}.")

        if self.__best_f_scaled is not None:
            self.__best_f_scaled.get_footer_comments(dest)
            dest(f"In summary {csv_scope(scope, KEY_BEST_F_SCALED)} describes"
                 " the best objective value reached ("
                 f"{csv_scope(scope, KEY_BEST_F)}) divided by the goal "
                 f"objective value ({csv_scope(scope, KEY_GOAL_F)}).")

        if self.__has_n_success:
            dest(f"{csv_scope(scope, KEY_N_SUCCESS)} is the number of runs "
                 "that reached goal objective value "
                 f"{csv_scope(scope, KEY_GOAL_F)}. Obviously, "
                 f"0<={csv_scope(scope, KEY_N_SUCCESS)}<="
                 f"{csv_scope(scope, KEY_N)}.")
        if self.__success_fes is not None:
            self.__success_fes.get_footer_comments(dest)
            dest(f"{csv_scope(scope, KEY_SUCCESS_FES)} offers statistics "
                 "about the number of FEs that the 0<="
                 f"{csv_scope(scope, KEY_N_SUCCESS)}<="
                 f"{csv_scope(scope, KEY_N)} successful runs needed to reach "
                 f"the goal objective value {csv_scope(scope, KEY_GOAL_F)}.")

        if self.__success_time_millis is not None:
            self.__success_fes.get_footer_comments(dest)
            dest(f"{csv_scope(scope, KEY_SUCCESS_TIME_MILLIS)} offers "
                 "statistics about the number of milliseconds of clock time "
                 f"that the 0<={csv_scope(scope, KEY_N_SUCCESS)}<="
                 f"{csv_scope(scope, KEY_N)} successful runs needed to reach "
                 f"the goal objective value {csv_scope(scope, KEY_GOAL_F)}.")

        if self.__has_ert_fes:
            dest(f"{csv_scope(scope, KEY_ERT_FES)} is the empirical estimate"
                 " of the number of FEs to solve the problem. It can be "
                 "approximated by dividing the sum of "
                 f"{csv_scope(scope, KEY_TOTAL_FES)} over all runs by the "
                 f"number {csv_scope(scope, KEY_N_SUCCESS)} of successful "
                 "runs.")

        if self.__has_ert_time_millis:
            dest(f"{csv_scope(scope, KEY_ERT_TIME_MILLIS)} is the empirical "
                 "estimate of the number of FEs to solve the problem. It can "
                 "be approximated by dividing the sum of "
                 f"{csv_scope(scope, KEY_TOTAL_TIME_MILLIS)} over all runs by"
                 f" the number {csv_scope(scope, KEY_N_SUCCESS)} of "
                 "successful runs.")

        if self.__max_fes is not None:
            self.__max_fes.get_footer_comments(dest)
            dest(f"In summary {csv_scope(scope, KEY_MAX_FES)} is"
                 f" {DESC_MAX_FES}.")
        if self.__max_time_millis is not None:
            self.__max_time_millis.get_footer_comments(dest)
            dest(f"In summary {csv_scope(scope, KEY_MAX_TIME_MILLIS)} is"
                 f" {DESC_MAX_TIME_MILLIS}.")

    def get_footer_bottom_comments(self, dest: Callable[[str], None]) -> None:
        """
        Get the footer bottom comments.

        :param dest: the destination to write to.
        """
        motipy_footer_bottom_comments(
            self, dest, ("The end statistics data is produced using module "
                         "moptipy.evaluation.end_statistics."))
        StatWriter.get_footer_bottom_comments(self.__best_f, dest)


class CsvReader:
    """A csv parser for end results."""

    def __init__(self, columns: dict[str, int]) -> None:
        """
        Create a CSV parser for :class:`EndResult`.

        :param columns: the columns
        """
        super().__init__()
        if not isinstance(columns, dict):
            raise type_error(columns, "columns", dict)

        #: the index of the algorithm column, if any
        self.__idx_algorithm: Final[int | None] = csv_column_or_none(
            columns, KEY_ALGORITHM)
        #: the index of the instance column, if any
        self.__idx_instance: Final[int | None] = csv_column_or_none(
            columns, KEY_INSTANCE)
        #: the index of the objective column, if any
        self.__idx_objective: Final[int | None] = csv_column_or_none(
            columns, KEY_OBJECTIVE_FUNCTION)
        #: the index of the encoding column, if any
        self.__idx_encoding: Final[int | None] = csv_column_or_none(
            columns, KEY_ENCODING)

        #: the index of the `N` column, i.e., where the number of runs is
        #: stored
        self.idx_n: Final[int] = csv_column(columns, KEY_N, True)

        n_key: Final[tuple[tuple[str, int]]] = ((KEY_N, self.idx_n), )
        #: the reader for the best-objective-value-reached statistics
        self.__best_f: Final[StatReader] = csv_select_scope(
            StatReader, columns, KEY_BEST_F, n_key)
        #: the reader for the last improvement FE statistics
        self.__life: Final[StatReader] = csv_select_scope(
            StatReader, columns, KEY_LAST_IMPROVEMENT_FE, n_key)
        #: the reader for the last improvement millisecond index statistics
        self.__lims: Final[StatReader] = csv_select_scope(
            StatReader, columns, KEY_LAST_IMPROVEMENT_TIME_MILLIS, n_key)
        #: the reader for the total FEs statistics
        self.__total_fes: Final[StatReader] = csv_select_scope(
            StatReader, columns, KEY_TOTAL_FES, n_key)
        #: the reader for the total milliseconds consumed statistics
        self.__total_ms: Final[StatReader] = csv_select_scope(
            StatReader, columns, KEY_TOTAL_TIME_MILLIS, n_key)

        #: the reader for the goal objective value statistics, if any
        self.__goal_f: Final[StatReader | None] = csv_select_scope_or_none(
            StatReader, columns, KEY_GOAL_F, n_key)
        #: the reader for the best-f / goal-f statistics, if any
        self.__best_f_scaled: Final[StatReader | None] = \
            csv_select_scope_or_none(
                StatReader, columns, KEY_BEST_F_SCALED, n_key)

        #: the index of the column where the number of successful runs is
        #: stored
        self.__idx_n_success: Final[int | None] = csv_column_or_none(
            columns, KEY_N_SUCCESS)
        succ_key: Final[tuple[tuple[str, int], ...]] = () \
            if self.__idx_n_success is None else (
            (KEY_N, self.__idx_n_success), )
        #: the reader for the success FE data, if any
        self.__success_fes: Final[StatReader | None] = \
            None if self.__idx_n_success is None else \
            csv_select_scope_or_none(
                StatReader, columns, KEY_SUCCESS_FES, succ_key)
        #: the reader for the success time milliseconds data, if any
        self.__success_time_millis: Final[StatReader | None] = \
            None if self.__idx_n_success is None else \
            csv_select_scope_or_none(
                StatReader, columns, KEY_SUCCESS_TIME_MILLIS, succ_key)

        #: the index of the expected FEs until success
        self.__idx_ert_fes: Final[int | None] = csv_column_or_none(
            columns, KEY_ERT_FES)
        #: the index of the expected milliseconds until success
        self.__idx_ert_time_millis: Final[int | None] = csv_column_or_none(
            columns, KEY_ERT_TIME_MILLIS)

        #: the columns with the maximum FE-based budget statistics
        self.__max_fes: Final[StatReader | None] = csv_select_scope_or_none(
            StatReader, columns, KEY_MAX_FES, n_key)
        #: the columns with the maximum time-based budget statistics
        self.__max_time_millis: Final[StatReader | None] = \
            csv_select_scope_or_none(
                StatReader, columns, KEY_MAX_TIME_MILLIS, n_key)

    def parse_row(self, data: list[str]) -> EndStatistics:
        """
        Parse a row of data.

        :param data: the data row
        :return: the end result statistics
        """
        return EndStatistics(
            algorithm=csv_str_or_none(data, self.__idx_algorithm),
            instance=csv_str_or_none(data, self.__idx_instance),
            objective=csv_str_or_none(data, self.__idx_objective),
            encoding=csv_str_or_none(data, self.__idx_encoding),
            n=int(data[self.idx_n]),
            best_f=self.__best_f.parse_row(data),
            last_improvement_fe=self.__life.parse_row(data),
            last_improvement_time_millis=self.__lims.parse_row(data),
            total_fes=self.__total_fes.parse_row(data),
            total_time_millis=self.__total_ms.parse_row(data),
            goal_f=StatReader.parse_optional_row(self.__goal_f, data),
            best_f_scaled=StatReader.parse_optional_row(
                self.__best_f_scaled, data),
            n_success=csv_val_or_none(data, self.__idx_n_success, int),
            success_fes=StatReader.parse_optional_row(
                self.__success_fes, data),
            success_time_millis=StatReader.parse_optional_row(
                self.__success_time_millis, data),
            ert_fes=csv_val_or_none(data, self.__idx_ert_fes, str_to_num),
            ert_time_millis=csv_val_or_none(
                data, self.__idx_ert_time_millis, str_to_num),
            max_fes=StatReader.parse_optional_row(self.__max_fes, data),
            max_time_millis=StatReader.parse_optional_row(
                self.__max_time_millis, data),
        )


@dataclass(frozen=True, init=False, order=False, eq=False)
class __PvEndStatistics(EndStatistics):
    """Aggregated end statistics."""

    pv: int | float

    def __init__(self, es: EndStatistics, pv: int | float):
        """
        Create the end statistics of an experiment-setup combination.

        :param es: the original end statistics object
        :param pv: the parameter value
        """
        super().__init__(
            es.algorithm, es.instance, es.objective, es.encoding, es.n,
            es.best_f, es.last_improvement_fe,
            es.last_improvement_time_millis, es.total_fes,
            es.total_time_millis, es.goal_f, es.best_f_scaled, es.n_success,
            es.success_fes, es.success_time_millis, es.ert_fes,
            es.ert_time_millis, es.max_fes, es.max_time_millis)
        if not isinstance(pv, (int | float)):
            raise type_error(pv, "pv", (int, float))
        if not isfinite(pv):
            raise ValueError(f"got pv={pv}")
        object.__setattr__(self, "pv", pv)

    def get_param_value(self) -> int | float:
        """
        Get the parameter value.

        :return: the parameter value
        """
        return self.pv


def aggregate_over_parameter(
        data: Iterable[EndResult],
        param_value: Callable[[EndResult], int | float],
        join_all_algorithms: bool = False,
        join_all_instances: bool = False,
        join_all_objectives: bool = False,
        join_all_encodings: bool = False) -> tuple[
        Callable[[EndStatistics], int | float], Iterable[EndStatistics]]:
    """
    Aggregate a stream of data into groups based on a parameter.

    :param data: the source data
    :param param_value: the function obtaining a parameter value
    :param join_all_algorithms: should the statistics be aggregated
        over all algorithms
    :param join_all_instances: should the statistics be aggregated
        over all algorithms
    :param join_all_objectives: should the statistics be aggregated over
        all objectives?
    :param join_all_encodings: should statistics be aggregated over all
        encodings
    """
    param_map: Final[dict[int | float, list[EndResult]]] = {}
    for er in data:
        pv = param_value(er)
        if not isinstance(pv, (int | float)):
            raise type_error(pv, f"param_value{er}", (int, float))
        if not isfinite(pv):
            raise ValueError(f"got {pv} = param({er})")
        if pv in param_map:
            param_map[pv].append(er)
        else:
            param_map[pv] = [er]
    if dict.__len__(param_map) <= 0:
        raise ValueError("Did not encounter any data.")

    stats: Final[list[EndStatistics]] = []
    for pv in sorted(param_map.keys()):
        def __make(es: EndStatistics, ppv=pv, ss=stats.append) -> None:
            ss(__PvEndStatistics(es, ppv))

        from_end_results(
            param_map[pv], cast(Callable[[EndStatistics], None], __make),
            join_all_algorithms, join_all_instances, join_all_objectives,
            join_all_encodings)
    return cast(Callable[[EndStatistics], int | float],
                __PvEndStatistics.get_param_value), tuple(stats)


# Run end-results to stat file if executed as script
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipy_argparser(
        __file__, "Build an end-results statistics CSV file.",
        "This program creates a CSV file with basic statistics on the "
        "end-of-run state of experiments conducted with moptipy. It "
        "therefore either parses a directory structure with log files "
        "(if src identifies a directory) or a end results CSV file (if"
        " src identifies a file). In the former case, the directory "
        "will follow the form 'algorithm/instance/log_file' with one "
        "log file per run. In the latter case, it will be a file "
        "generated by the end_results.py tool of moptipy. The output "
        "of this tool is a CSV file where the columns are separated by"
        " ';' and the rows contain the statistics.")
    def_src: str = "./evaluation/end_results.txt"
    if not os.path.isfile(def_src):
        def_src = "./results"
    parser.add_argument(
        "source", nargs="?", default=def_src,
        help="either the directory with moptipy log files or the path to the "
             "end-results CSV file", type=Path)
    parser.add_argument(
        "dest", type=Path, nargs="?",
        default="./evaluation/end_statistics.txt",
        help="the path to the end results statistics CSV file to be created")
    parser.add_argument(
        "--join_algorithms",
        help="compute statistics over all algorithms, i.e., the statistics"
             " are not separated by algorithm but all algorithms are treated "
             "as one", action="store_true")
    parser.add_argument(
        "--join_instances",
        help="compute statistics over all instances, i.e., the statistics"
             " are not separated by instance but all instances are treated "
             "as one", action="store_true")
    parser.add_argument(
        "--join_objectives",
        help="compute statistics over all objective functions, i.e., the "
             "statistics are not separated by objective functions but all "
             "objectives functions are treated as one", action="store_true")
    parser.add_argument(
        "--join_encodings",
        help="compute statistics over all encodings, i.e., the statistics"
             " are not separated by encodings but all encodings are treated "
             "as one", action="store_true")
    args: Final[argparse.Namespace] = parser.parse_args()

    src_path: Final[Path] = args.source
    end_results: Final[list[EndResult]] = []
    if src_path.is_file():
        logger(f"{src_path!r} identifies as file, load as end-results csv")
        end_results_from_csv(src_path, end_results.append)
    else:
        logger(f"{src_path!r} identifies as directory, load it as log files")
        end_results_from_logs(src_path, end_results.append)

    end_stats: Final[list[EndStatistics]] = []
    from_end_results(
        source=end_results, consumer=end_stats.append,
        join_all_algorithms=args.join_algorithms,
        join_all_instances=args.join_instances,
        join_all_objectives=args.join_objectives,
        join_all_encodings=args.join_encodings)
    to_csv(end_stats, args.dest)



"""
An operator trying to swap a given number of elements in a permutation.

This operator is similar to
:mod:`~moptipy.operators.permutations.op1_swap_exactly_n`
in that it tries to swap a fixed number of elements in a permutation. It
implements the :class:`~moptipy.api.operators.Op1WithStepSize` interface.
Different from :mod:`~moptipy.operators.permutations.op1_swap_exactly_n`,
however, it does so less strictly. It applies a simple best-effort approach
and if that does not work out, then so be it. It is therefore faster, but
adheres less strictly to the give `step_size`.

This operator will always swap the right number of elements on normal
permutations. On permutations with repetitions, it enforces the number of
swaps less strongly compared to
:mod:`~moptipy.operators.permutations.op1_swap_exactly_n`, but it will be
faster either way.
"""
from typing import Final

import numpy as np
from numpy.random import Generator
from pycommons.types import type_error

from moptipy.api.operators import Op1WithStepSize
from moptipy.operators.tools import exponential_step_size
from moptipy.spaces.permutations import Permutations
from moptipy.utils.nputils import DEFAULT_INT, fill_in_canonical_permutation


# Temporary fix for https://github.com/numba/numba/issues/9103
def swap_try_n(random: Generator, dest: np.ndarray, x: np.ndarray,
               step_size: float, indices: np.ndarray) -> None:
    """
    Copy `x` into `dest` and then swap several different values.

    :param random: the random number generator
    :param dest: the array to receive the modified copy of `x`
    :param x: the existing point in the search space
    :param step_size: the number of elements to swap
    :param indices: the array with indices

    >>> xx = np.array(range(10), int)
    >>> out = np.empty(len(xx), xx.dtype)
    >>> idxs = np.array(range(len(xx)), int)
    >>> gen = np.random.default_rng(10)
    >>> swap_try_n(gen, out, xx, 0.1, idxs)
    >>> print(out)
    [0 1 2 3 4 5 6 8 7 9]
    >>> swap_try_n(gen, out, xx, 0.1, idxs)
    >>> print(out)
    [0 2 1 3 4 5 6 7 8 9]
    >>> swap_try_n(gen, out, xx, 1.0, idxs)
    >>> print(out)
    [3 7 4 5 8 6 9 0 1 2]
    """
    dest[:] = x[:]  # First, we copy `x` to `dest`.
    remaining: int = len(dest)  # Get the length of `dest`.
# Compute the real step size based on the length of the permutation.
    steps: int = exponential_step_size(step_size, 2, remaining)

    ii: int = random.integers(0, remaining)  # Select first random index.
    i1 = indices[ii]  # Get the actual index.
    remaining = remaining - 1  # There is one less remaining index.
    indices[remaining], indices[ii] = i1, indices[remaining]

    last = first = dest[i1]  # Get the value at the first index.
    continue_after: bool = True  # True -> loop at least once.
    while continue_after:  # Repeat until we should stop
        steps = steps - 1
        continue_after = (steps > 1) and (remaining > 1)
        tested: int = remaining  # the indices that we can test.
        while True:  # Loop forever until eligible element found.
            ii = random.integers(0, tested)  # Get a new random index.
            i2 = indices[ii]  # Get the actual index.
            current = dest[i2]  # Get the value at the new index.
            if tested <= 1:  # Are all remaining elements same?
                continue_after = False  # If yes, then we quit.
                break
            if (current != last) and (
                    continue_after or (current != first)):
                remaining = remaining - 1
                indices[remaining], indices[ii] = \
                    i2, indices[remaining]
                break  # to stop, then it must be != first value.
            tested = tested - 1  # Now there is one fewer index.
            indices[tested], indices[ii] = i2, indices[tested]
        dest[i1] = last = current  # Store value for from i2 at i1.
        i1 = i2  # Update i1 to now point to cell of i2.
    dest[i1] = first  # Finally, store first element back at end.


class Op1SwapTryN(Op1WithStepSize):
    """An operator trying to swap a given number of elements."""

    def op1(self, random: Generator, dest: np.ndarray, x: np.ndarray,
            step_size: float = 0.0) -> None:
        """
        Copy `x` into `dest` and then swap several different values.

        :param random: the random number generator
        :param dest: the array to receive the modified copy of `x`
        :param x: the existing point in the search space
        :param step_size: the number of elements to swap
        """
        swap_try_n(random, dest, x, step_size, self.__indices)

    def __init__(self, perm: Permutations) -> None:
        """
        Initialize the operator.

        :param perm: the base permutation
        """
        super().__init__()
        if not isinstance(perm, Permutations):
            raise type_error(perm, "perm", Permutations)
        #: the valid indices
        self.__indices: Final[np.ndarray] = np.empty(
            perm.dimension, DEFAULT_INT)

    def initialize(self) -> None:
        """Initialize this operator."""
        super().initialize()
        fill_in_canonical_permutation(self.__indices)

    def __str__(self) -> str:
        """
        Get the name of this unary operator.

        :returns: "swaptn", the name of this operator
        :retval "swaptn": always
        """
        return "swaptn"



"""Here we provide some tools for implementing FFA."""

from collections import Counter
from io import StringIO
from math import nan
from typing import Callable, Final

import numpy as np
from pycommons.io.csv import CSV_SEPARATOR
from pycommons.math.int_math import try_int
from pycommons.types import type_error

from moptipy.api.objective import Objective
from moptipy.api.process import Process
from moptipy.utils.nputils import DEFAULT_INT

#: the log section for the frequency table
H_LOG_SECTION: Final[str] = "H"

#: The difference between upper- and lower bound at which we switch from
#: using arrays as backing store for the frequency table H to maps.
SWITCH_TO_MAP_RANGE: Final[int] = 67_108_864


def create_h(objective: Objective) -> tuple[
        np.ndarray | Counter[int | float], int]:
    """
    Create a data structure for recording frequencies.

    If our objective function always returns integer values and its range is
    not too large, then we can use a :class:`numpy.ndarray` as a backing
    datastructure for the `H`-map used in FFA. If the objective function may
    sometimes return `float` values or has a wide range, it will be better
    (or even necessary) to use :class:`Counter` instead.

    This function here makes the decision of what backing datastructure to
    use.

    Additionally, it may be the lower or even upper bound of an integer
    objective function could be negative. To avoid dealing with negative
    indices, we would then add an offset to each objective value before
    accessing it. This function will return an appropriate offset as well.

    :param objective: the objective for which the data structure should be
        created
    :return: a tuple of the `H`-data structure as well as an offset to be
        added to all objective values.
    """
    if not isinstance(objective, Objective):
        raise type_error(objective, "objective", Objective)
    if objective.is_always_integer():
        lb: Final[int | float] = objective.lower_bound()
        ub: Final[int | float] = objective.upper_bound()
        if isinstance(ub, int) and isinstance(lb, int) \
                and ((ub - lb) <= SWITCH_TO_MAP_RANGE):
            return np.zeros(ub - lb + 1, DEFAULT_INT), -lb
    return Counter(), 0


def clear_h(h: np.ndarray | Counter[int | float]) -> None:
    """
    Clear a H-Table.

    :param h: the H-table to clear.
    """
    if isinstance(h, np.ndarray):
        h.fill(0)
    elif isinstance(h, dict):
        h.clear()
    else:
        raise type_error(h, "h", (np.ndarray, Counter))


def h_to_str(
        h: np.ndarray | Counter[int | float],
        offset: int) -> str:
    """
    Convert a `H`-table to a string.

    :param h: the `H`-table, as created by :func:`create_h`.
    :param offset: the offset, as computed by :func:`create_h`.

    >>> hl = np.array([0, 0, 1, 7, 4, 0, 0, 9, 0])
    >>> h_to_str(hl, 0)
    '2;1;;7;;4;7;9'
    >>> h_to_str(hl, -1)
    '3;1;;7;;4;8;9'
    >>> hd = Counter((1, 1, 1, 1, 1, 4, 4, 4, 4, 4, 4, 4, 3, 3, 3, 3,
    ...               3, 3, 2, 2, 2, 2, 2, 2, 2, 2, 2))
    >>> h_to_str(hd, 0)
    '1;5;;9;;6;;7'
    >>> try:
    ...     hd = {1: 0}
    ...     h_to_str(hd, 0)
    ... except ValueError as ve:
    ...     print(ve)
    empty H table?
    >>> hx = np.zeros(100, int)
    >>> hx[10] = 4
    >>> hx[12] = 234
    >>> hx[89] = 111
    >>> hx[90] = 1
    >>> hx[45] = 2314
    >>> h_to_str(hx, 0)
    '10;4;12;234;45;2314;89;111;;1'
    """
    if not isinstance(h, np.ndarray | dict):
        raise type_error(h, "h", (np.ndarray, Counter))
    if not isinstance(offset, int):
        raise type_error(offset, "offset", int)
    csep: Final[str] = CSV_SEPARATOR
    sep: str = ""

    with StringIO() as out:
        write: Callable[[str], int] = out.write  # fast call
        old_index: int | float = nan

        # We iterate over the whole `H` table.
        # If the table is an np.array, then we just use the indices to
        # iterate directly. If it is a `Counter`, then we iterate over
        # its keys. If there are not too many keys, then we sort them
        # first.
        for i in range(len(h)) if isinstance(h, np.ndarray) else (sorted(
                h.keys()) if len(h) <= SWITCH_TO_MAP_RANGE else h.keys()):
            value: int | float = h[i]  # type: ignore
            if value <= 0:  # skip over values that have never been hit
                continue
            write(sep)  # write separator (nothing upon first call)
            sep = csep  # now the next call will write ;
            use_index: int | float = i - offset  # subtract the offset
            if isinstance(use_index, float):  # if it's float, try to convert
                use_index = try_int(use_index)
            if (use_index - 1) != old_index:  # we skip if current = old + 1
                write(str(use_index))
            old_index = use_index  # step the index
            write(csep)  # write separator to frequency counter
            write(str(value))  # write the frequency
        res: Final[str] = out.getvalue()  # get the final string

    if str.__len__(res) <= 0:
        raise ValueError("empty H table?")
    return res


def log_h(process: Process, h: np.ndarray | Counter[int | float],
          offset: int) -> None:
    """
    Convert a frequency table `H` to a string and log it to a process.

    The frequency table is logged as a single line of text into a section
    `H` delimited by the lines `BEGIN_H` and `END_H`. The line consists
    of `2*n` semicolon separated values. Each such value pair consists of
    an objective value `y` and its observed frequency `H[y]`. The former is
    either an integer or a float and the latter is an integer.
    If two consecutive objective values take the form 'f` and `f+1`, then
    the second one will be omitted, and so on.

    :param process: the process
    :param h: the `H`-table, as created by :func:`create_h`.
    :param offset: the offset, as computed by :func:`create_h`.
    """
    if not isinstance(process, Process):
        raise type_error(process, "process", Process)
    if process.has_log():
        process.add_log_section(H_LOG_SECTION, h_to_str(h, offset))



"""
Approximate the ECDF over the ERT to reach certain goals.

The empirical cumulative distribution function (ECDF, see
:mod:`~moptipy.evaluation.ecdf`) is a function that shows the fraction of runs
that were successful in attaining a certain goal objective value over the
time. The (empirically estimated) Expected Running Time (ERT, see
:mod:`~moptipy.evaluation.ert`) is a function that tries to give an estimate
how long a given algorithm setup will need (y-axis) to achieve given solution
qualities (x-axis). It uses a set of runs of the algorithm on the problem to
make this estimate under the assumption of independent restarts.

Now in the ERT-ECDF we combine both concepts to join several different
optimization problems or problem instances into one plot. The goal becomes
"solving the problem". For each problem instance, we compute the ERT, i.e.,
estimate how long a given algorithm will need to reach the goal. This becomes
the time axis. Over this time axis, the ERT-ECDF displays the fraction of
instances that were solved.

1. Thomas Weise, Zhize Wu, Xinlu Li, and Yan Chen. Frequency Fitness
   Assignment: Making Optimization Algorithms Invariant under Bijective
   Transformations of the Objective Function Value. *IEEE Transactions on
   Evolutionary Computation* 25(2):307-319. April 2021. Preprint available at
   arXiv:2001.01416v5 [cs.NE] 15 Oct 2020. http://arxiv.org/abs/2001.01416.
   doi: https://doi.org/10.1109/TEVC.2020.3032090
"""

from dataclasses import dataclass
from typing import Any, Callable, Iterable

from moptipy.evaluation.ecdf import Ecdf
from moptipy.evaluation.ert import compute_single_ert
from moptipy.evaluation.progress import Progress


@dataclass(frozen=True, init=False, order=False, eq=False)
class ErtEcdf(Ecdf):
    """The ERT-ECDF."""

    def time_label(self) -> str:
        """
        Get the time axis label.

        :return: the time key
        """
        return f"ERT\u2009[{self.time_unit}]"

    def _time_key(self) -> str:
        """
        Get the time key.

        :return: the time key
        """
        return f"ert[{super()._time_key()}]"

    @staticmethod
    def _compute_times(source: list[Progress],
                       goal: int | float) -> list[float]:
        """
        Compute the times for the given goals.

        Warning: `source` must only contain progress objects that contain
        monotonously improving points. It must not contain runs that may get
        worse over time.

        :param source: the source array
        :param goal: the goal value
        :return: a list of times
        """
        return [compute_single_ert(source, goal)]

    # noinspection PyUnusedLocal
    @staticmethod
    def _get_div(n: int, n_insts: int) -> int:
        """
        Get the divisor.

        :param n: the number of runs
        :param n_insts: the number of instances
        :return: the divisor
        """
        del n
        return n_insts


def create(source: Iterable[Progress],
           goal_f: int | float | Callable[[str], int | float] | None = None,
           use_default_goal_f: bool = True) -> Ecdf:
    """
    Create one single Ert-Ecdf record from an iterable of Progress records.

    :param source: the set of progress instances
    :param goal_f: the goal objective value
    :param use_default_goal_f: should we use the default lower bounds as
        goals?
    :return: the Ert-Ecdf record
    """
    return ErtEcdf._create(source, goal_f, use_default_goal_f)


def from_progresses(
        source: Iterable[Progress], consumer: Callable[[Ecdf], Any],
        f_goal: int | float | Callable[[str], int | float]
                    | Iterable[int | float | Callable] | None = None,
        join_all_algorithms: bool = False,
        join_all_objectives: bool = False,
        join_all_encodings: bool = False) -> None:
    """
    Compute one or multiple Ert-ECDFs from a stream of end results.

    :param source: the set of progress instances
    :param f_goal: one or multiple goal values
    :param consumer: the destination to which the new records will be
        passed, can be the `append` method of a :class:`list`
    :param join_all_algorithms: should the Ecdf be aggregated over all
        algorithms
    :param join_all_objectives: should the Ecdf be aggregated over all
        objective functions
    :param join_all_encodings: should the Ecdf be aggregated over all
        encodings
    """
    return ErtEcdf._from_progresses(
        source, consumer, f_goal, join_all_algorithms,
        join_all_objectives, join_all_encodings)



"""Perform restarts of an algorithm that terminates too early."""
from typing import Callable, Final, TypeVar, cast

from pycommons.types import type_error

from moptipy.api.algorithm import Algorithm
from moptipy.api.mo_algorithm import MOAlgorithm
from moptipy.api.mo_process import MOProcess
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection


class __RestartAlgorithm(Algorithm):
    """A wrapper around an existing algorithm."""

    def __init__(self, algorithm: Algorithm) -> None:
        """
        Create the algorithm wrapper.

        :param algorithm: the algorithm to wrap
        """
        super().__init__()
        #: the algorithm
        self._algo: Final[Algorithm] = algorithm
        self.initialize = algorithm.initialize  # type: ignore # fast call

    def __str__(self) -> str:
        """
        Get the string representation of this algorithm.

        :returns: `rs_` + sub-algorithm name
        """
        return f"rs_{self._algo}"

    def solve(self, process: Process) -> None:
        """
        Solve a single-objective problem.

        :param process: the process
        """
        st: Final[Callable[[], bool]] = process.should_terminate
        rst: Final[Callable[[], None]] = self._algo.initialize
        sv: Final[Callable[[Process], None]] = self._algo.solve
        while not st():
            rst()
            sv(process)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Store the parameters of this algorithm wrapper to the logger.

        :param logger: the logger
        """
        super().log_parameters_to(logger)
        with logger.scope("a") as scope:
            self._algo.log_parameters_to(scope)


class __RestartMOAlgorithm(__RestartAlgorithm, MOAlgorithm):
    """A wrapper around an existing multi-objective algorithm."""

    def solve_mo(self, process: MOProcess) -> None:
        """
        Solve a multi-objective problem.

        :param process: the process
        """
        st: Final[Callable[[], bool]] = process.should_terminate
        rst: Final[Callable[[], None]] = self.initialize
        sv: Final[Callable[[MOProcess], None]] = cast(
            MOAlgorithm, self._algo).solve_mo
        while not st():
            rst()
            sv(process)

    def __str__(self) -> str:
        """
        Get the string representation of this algorithm.

        :returns: `mors_` + sub-algorithm name
        """
        return f"mors_{self._algo}"


#: the type variable for single- and multi-objective algorithms.
T = TypeVar("T", Algorithm, MOAlgorithm)


def restarts(algorithm: T) -> T:
    """
    Perform restarts of an algorithm until the termination criterion is met.

    :param algorithm: the algorithm
    """
    if isinstance(algorithm, MOAlgorithm):
        return __RestartMOAlgorithm(algorithm)
    if isinstance(algorithm, Algorithm):
        return __RestartAlgorithm(algorithm)
    raise type_error(algorithm, "algorithm", (Algorithm, MOAlgorithm))



"""
A simple implementation of a Plant Propagation Algorithm (PPA).

This is a simple implementation of the Plant Propagation Algorithm, PPA for
short, with some tweaks and modifications.
Our PPA implementation works as follows:

1. It starts with a set of :attr:`~moptipy.algorithms.so.ppa.PPA.m` randomly
   sampled solutions in the list `lst`. Each solution `x` is evaluated and its
   objective value `f(x)` is remembered.
2. In the main loop...

   a. First, the range `[fmin,fmax]` of the objective values of the first
      :attr:`~moptipy.algorithms.so.ppa.PPA.m` solutions in `lst` is
      determined. We set `frange = fmax - fmin`, where `fmax` is the largest
      objective value of any of the first `m` solutions in `lst` and `fmin`
      is the smallest one. If `frange > 0`, then the fitness `z(x)` of each
      element be `(f(x) - fmin) / frange`. Otherwise, i.e., if all solutions
      in `lst` have the same objective value, we set `z(x)` to be a random
      number uniformly distributed in `[0,1)` and drawn separately for each
      solution.
   b. For each of the first :attr:`~moptipy.algorithms.so.ppa.PPA.m` solutions
      `x` in `lst`, we create `1 + int(nmax * r * (1 - z(x)))` offspring,
      where :attr:`~moptipy.algorithms.so.ppa.PPA.nmax` is the maximum number
      of offspring per solution and `r` be again an independently drawn random
      number uniformly distributed in `[0,1)`. In other words, solutions with
      a fitness close to zero will produce more offspring. If the solutions in
      the list `lst` have different objective values, then this means that
      better solutions produce more offsprings.
      Each such offspring is the result of the application of a unary operator
      with step size, i.e., an instance of
      :class:`~moptipy.api.operators.Op1WithStepSize`. The step size is set to
      `r * max_step * z(x)`, where `r` again is a freshly and independently
      drawn random number uniformly distributed in `[0,1)`. This means that
      better solutions are modified with smaller step sizes and worse
      solutions are modified more strongly.
      :attr:`~moptipy.algorithms.so.ppa.PPA.max_step` is a parameter of the
      algorithm that determines the maximum permissible step size. It is
      always from the interval `[0,1]`.
      Examples for such operators are given in
      :mod:`~moptipy.operators.permutations.op1_swap_exactly_n`,
      :mod:`~moptipy.operators.permutations.op1_swap_try_n`, or
      :mod:`~moptipy.operators.bitstrings.op1_flip_m`.
      The new solutions are appended into `lst` and their objective values are
      computed.
   c. The list is then sorted by objective values in ascending order, meaning
      that the best solutions are up front.

The main differences between this procedure and the "standard-PPA" are as
follows:

A. The algorithm is implemented for minimization and all equations are
   modified accordingly.
B. After normalizing the objective values in the population, the `tanh`-based
   scaling is *not* applied.
   Instead, the normalized objective values, where `0` is best and `1` is
   worst, are used directly to determine the number of offspring per record
   and the step length.
C. The fitness of a record equals its normalized objective value
   (in `[0, 1]`), unless all records have the same objective value, in which
   case the fitness of each record is set to a random number uniformly
   distributed in `[0, 1)`.
   If all elements in the population have the same objective value,
   normalizing is not possible as it would lead to a division by zero.
   One could use a constant value, say `0.5`, in this case, but there is no
   guarantee that this would be a good choice.
   We therefore use random values from `[0, 1)` instead.
   These may sometimes be suitable, sometimes not.
   But at least they likely are not *always* a bad choice, which might happen
   in some scenarios with `0.5` or any other constant.
D. The decisions regarding the number of offspring per selected record and the
   step-width of the search moves are made only based on this fitness (and,
   again, not on the `tanh` scaling which is not used).
   Since we normally do not know the characteristics of the objective function
   in advance, I think that we also often do not know whether a `tanh` scaling
   (that emphasizes objective values close to the best and close to the worst)
   is necessary or a good idea.
   It could be good in some cases, but it might as well be a bad choice in
   others.
   For now, I have thus not implemented this and just use the raw normalized
   objective values.
E. As unary operators, we employ instances of the class
   :class:`~moptipy.api.operators.Op1WithStepSize`, which provides a unary
   operator with a step size between `0` (smallest possible modification) to
   `1` (largest possible modification) and will scale appropriately between
   the two extremes.
   Often, instances of this class will determine the number or magnitude of
   changes based on an exponential scaling (see
   :func:`~moptipy.operators.tools.exponential_step_size`) of the step length.
   The idea is that small step sizes should be emphasized and that really big
   step sizes are often rarely needed.
   This thus effectively takes the place of the `tanh` scaling.
F. Maximum step lengths, i.e., the parameter
   :attr:`~moptipy.algorithms.so.ppa.PPA.max_step`, are not always explicitly
   used in some of the papers.

In order to understand the behavior of the algorithm, consider the following
case. Assume that we set the maximum number
(:attr:`~moptipy.algorithms.so.ppa.PPA.nmax`) of offspring per solution to `1`
and the number :attr:`~moptipy.algorithms.so.ppa.PPA.m` of solutions to
survive selection to `1` as well. In this case, the PPA has exactly the same
"population structure" as the Randomized Local Search
(:mod:`~moptipy.algorithms.so.rls`), namely it preserves the best-so-far
solution and generates one new solution in each step, which then competes with
that best-so-far solution. The two algorithms then only differ in their search
operator: The step-length of the unary operator used in PPA depends on the
relative objective value and the one of the RLS does not. However, there are
many situations where the two could still be equivalent. For example, if the
current and new solution have different objective values, normalizing the
objective value will mean that the best of the two has normalized objective
value "0". This equates to the shortest possible step length. In this case,
for example, the step-length based operator
:mod:`~moptipy.operators.permutations.op1_swap_try_n` behaves exactly like the
:mod:`~moptipy.operators.permutations.op1_swap2` operator and the step-length
based :mod:`~moptipy.operators.bitstrings.op1_flip_m` operator behaves like
the :mod:`~moptipy.operators.bitstrings.op1_flip1`.
Of course, if both the current and the new solution have the same objective
value, then we use a random number from `[0,1)` as normalized objective value,
so the operators would not behave the same. Then again, one could set the
maximum step length :attr:`~moptipy.algorithms.so.ppa.PPA.max_step` to `0`.
In this case, the step length is always zero and most of our step-length based
operations will behave like fixed small step-length based counterparts, as
mentioned above. So in other words, if we set both
:attr:`~moptipy.algorithms.so.ppa.PPA.m` and
:attr:`~moptipy.algorithms.so.ppa.PPA.nmax` to `1` and set
:attr:`~moptipy.algorithms.so.ppa.PPA.max_step` to `0`, our PPA behaves like
:mod:`~moptipy.algorithms.so.rls` (if the search operators are appropriately
chosen).

Now :mod:`~moptipy.algorithms.so.rls` is also known as the (1+1) EA
and indeed, it is a special case of the (mu+lambda) EA implemented in
:mod:`~moptipy.algorithms.so.ea`.
I think with some appropriate settings of the parameter, we can probably
construct some setups of both algorithms with larger populations that should
be equivalent or close-to-equivalent in the big picture.

Below, you can find references on the PPA.

1. Abdellah Salhi and Eric Serafin Fraga. Nature-Inspired Optimisation
   Approaches and the New Plant Propagation Algorithm. *Proceeding of the
   International Conference on Numerical Analysis and Optimization
   (ICeMATH'2011),* June 6-8, 2011, Yogyakarta, Indonesia, volume 1,
   pages K2-1--K2-8. ISBN: 978-602-98919-1-1.
   https://doi.org/10.13140/2.1.3262.0806.
   https://repository.essex.ac.uk/9974/1/paper.pdf.
2. Misha Paauw and Daan van den Berg. Paintings, Polygons and Plant
   Propagation. In Anikó Ekárt, Antonios Liapis, and María Luz Castro Pena,
   editors, *Proceedings of the 8th International Conference on Computational
   Intelligence in Music, Sound, Art and Design (EvoMUSART'19, Part of
   EvoStar)*, April 24-26, 2019, Leipzig, Germany, Lecture Notes in Computer
   Science (LNCS), volume 11453, pages 84-97. ISBN: 978-3-030-16666-3. Cham,
   Switzerland: Springer. https://doi.org/10.1007/978-3-030-16667-0_6.
   https://www.researchgate.net/publication/332328080.
3. Muhammad Sulaiman, Abdellah Salhi, Eric Serafin Fraga, Wali Khan Mashwa,
   and Muhammad M. Rashi. A Novel Plant Propagation Algorithm: Modifications
   and Implementation. *Science International (Lahore)* 28(1):201-209, #2330,
   January/February 2016. http://www.sci-int.com/pdf/4066579081%20a%20201-\
209%20PPA%20Science%20international_Wali.pdf.
   https://arxiv.org/pdf/1412.4290.pdf
4. Hussein Fouad Almazini, Salah Mortada, Hassan Fouad Abbas Al-Mazini, Hayder
   Naser Khraibet AL-Behadili, and Jawad Alkenani. Improved Discrete Plant
   Propagation Algorithm for Solving the Traveling Salesman Problem. *IAES
   International Journal of Artificial Intelligence (IJ-AI)* 11(1):13-22.
   March 2022. http://doi.org/10.11591/ijai.v11.i1.pp13-22.
   https://www.researchgate.net/publication/357484222.
5. Birsen İrem Selamoğlu and Abdellah Salhi. The Plant Propagation Algorithm
   for Discrete Optimisation: The Case of the Travelling Salesman Problem.
   In Xin-She Yang, editor, *Nature-Inspired Computation in Engineering,*
   pages 43-61. Studies in Computational Intelligence (SCI), Volume 637.
   March 2016. Cham, Switzerland: Springer.
   https://doi.org/10.1007/978-3-319-30235-5_3.
   https://www.researchgate.net/publication/299286896.
6. Marleen de Jonge and Daan van den Berg. Parameter Sensitivity Patterns in
   the Plant Propagation Algorithm. In Juan Julián Merelo Guervós,
   Jonathan M. Garibaldi, Christian Wagner, Thomas Bäck, Kurosh Madani, and
   Kevin Warwick, editors, *Proceedings of the 12th International Joint
   Conference on Computational Intelligence* (IJCCI'20), November 2-4, 2020,
   Budapest, Hungary, pages 92-99. Setúbal, Portugal: SciTePress.
   https://doi.org/10.5220/0010134300920099.
   https://www.researchgate.net/publication/346829569.
7. Ege de Bruin. Escaping Local Optima by Preferring Rarity with the
   Frequency Fitness Assignment. Master's Thesis at Vrije Universiteit
   Amsterdam, Amsterdam, the Netherlands. 2022.
8. Wouter Vrielink and Daan van den Berg. Parameter control for the Plant
   Propagation Algorithm. In Antonio M. Mora and Anna Isabel Esparcia-Alcázar,
   editors, *Late-Breaking Abstracts of EvoStar'21*, April 7-9, 2021, online
   conference. https://arxiv.org/pdf/2106.11804.pdf.
   https://www.researchgate.net/publication/350328314.
9. Levi Koppenhol, Nielis Brouwer, Danny Dijkzeul, Iris Pijning, Joeri
   Sleegers, and Daan van den Berg. Exactly Characterizable Parameter Settings
   in a Crossoverless Evolutionary Algorithm. In Jonathan E. Fieldsend and
   Markus Wagner, editors, Genetic and Evolutionary Computation Conference
   (GECCO'22) Companion Volume, July 9-13, 2022, Boston, MA, USA,
   pages 1640-1649. New York, NY, USA: ACM.
   https://doi.org/10.1145/3520304.3533968.
   https://www.researchgate.net/publication/362120506.
"""
from math import isfinite
from typing import Callable, Final, cast

from numpy.random import Generator
from pycommons.types import check_int_range, type_error

from moptipy.algorithms.so.record import Record
from moptipy.api.algorithm import Algorithm1
from moptipy.api.operators import Op0, Op1WithStepSize
from moptipy.api.process import Process
from moptipy.utils.logger import KeyValueLogSection
from moptipy.utils.strings import PART_SEPARATOR, num_to_str_for_name


# start book
class PPA(Algorithm1):
    """The Plant Propagation Algorithm (PPA)."""

    def solve(self, process: Process) -> None:
        """
        Apply the PPA to an optimization problem.

        :param process: the black-box process object
        """
        m: Final[int] = self.m  # m: the number of best solutions kept
        nmax: Final[int] = self.nmax  # maximum offspring per solution
        list_len: Final[int] = (nmax + 1) * m
        # initialization of some variables omitted in book for brevity
        # end book
        random: Final[Generator] = process.get_random()  # random gen
        create: Final[Callable] = process.create  # create x container
        evaluate: Final[Callable] = process.evaluate  # the objective
        op0: Final[Callable] = self.op0.op0  # the nullary operator
        op1: Final[Callable] = cast(Op1WithStepSize,
                                    self.op1).op1  # the unary operator
        should_terminate: Final[Callable] = process.should_terminate
        r01: Final[Callable[[], float]] = cast(  # random floats
            Callable[[], float], random.random)
        max_step: Final[float] = self.max_step
        # start book
        # create list of m random records and enough empty records
        lst: Final[list] = [None] * list_len  # pre-allocate list
        f: int | float = 0  # variable to hold objective values
        for i in range(list_len):  # fill list of size m*nmax
            x = create()  # by creating point in search space
            if i < m:  # only the first m records are initialized by
                op0(random, x)  # applying nullary operator = randomize
                if should_terminate():  # should we quit?
                    return   # computational budget exhausted -> quit
                f = evaluate(x)  # continue? ok, evaluate new solution
            lst[i] = Record(x, f)  # create and store record

        it: int = 0  # the iteration counter
        while True:  # lst: keep 0..mu-1, overwrite mu..mu+lambda-1
            it = it + 1  # step iteration counter
            fmin = fmax = lst[0].f  # get range of objective values
            for i in range(m):  # iterate over selected individuals
                fval = lst[i].f  # get objective value
                if fval < fmin:  # is it less than minimum?
                    fmin = fval  # yes -> update the minimum
                elif fval > fmax:  # no! is it more than maximum then?
                    fmax = fval  # yes -> update maximum
            frange = fmax - fmin  # compute the range of objective
            all_same = (not isfinite(frange)) or (frange <= 0.0)
            total = m  # the total population length (so far: m)
            for i in range(m):  # generate offspring for each survivor
                rec = lst[i]  # get parent record
                fit = r01() if all_same else ((rec.f - fmin) / frange)
                x = rec.x  # the parent x
                for _ in range(1 + int((1.0 - fit) * r01() * nmax)):
                    if should_terminate():  # should we quit?
                        return  # yes - then return
                    dest = lst[total]  # get next destination record
                    total = total + 1  # remember we have now one more
                    dest.it = it  # set iteration counter
                    op1(random, dest.x, x, fit * max_step * r01())
                    dest.f = evaluate(dest.x)  # evaluate new point
            ls = lst[0:total]  # get sub-list of elements in population
            ls.sort()  # sort these used elements
            lst[0:total] = ls  # write the sorted sub-list back
# end book

    def __init__(self, op0: Op0, op1: Op1WithStepSize, m: int = 30,
                 nmax: int = 5, max_step: float = 0.3,
                 name: str = "ppa") -> None:
        """
        Create the Plant Propagation Algorithm (PPA).

        :param op0: the nullary search operator
        :param op1: the unary search operator
        :param m: the number of best solutions to survive in each generation
        :param nmax: the maximum number of offspring per solution
        :param name: the base name of the algorithm
        """
        if not isinstance(op1, Op1WithStepSize):
            raise type_error(op1, "op1", Op1WithStepSize)

        #: the number of records to survive in each generation
        self.m: Final[int] = check_int_range(m, "m", 1, 1_000_000)
        #: the maximum number of offsprings per solution per iteration
        self.nmax: Final[int] = check_int_range(
            nmax, "nmax", 1, 1_000_000)
        if not isinstance(max_step, float):
            raise type_error(max_step, "max_step", float)
        if (not isfinite(max_step)) or (max_step < 0.0) or (max_step > 1.0):
            raise ValueError(f"max_step={max_step}, but must be in [0,1].")
        #: the maximum step length
        self.max_step: Final[float] = max_step

        name = f"{name}{PART_SEPARATOR}{m}{PART_SEPARATOR}{nmax}"
        if max_step != 1.0:
            name = f"{name}{PART_SEPARATOR}{num_to_str_for_name(max_step)}"
        super().__init__(name, op0, op1)

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        logger.key_value("m", self.m)
        logger.key_value("nmax", self.nmax)
        logger.key_value("maxStep", self.max_step)



"""
Objects embodying the progress of a run over time.

An instance of :class:`Progress` holds one :attr:`~Progress.time` vector and
an objective value (:attr:`~Progress.f`) vector. The time dimension (stored in
:attr:`~Progress.time_unit`) can either be in FEs or in milliseconds and the
objective value dimension (stored in :attr:`~Progress.f_name`) can be raw
objective values, standardized objective values, or normalized objective
values.
The two vectors together thus describe how a run of an optimization algorithm
improves the objective value over time.
"""
from dataclasses import dataclass
from math import inf, isfinite
from typing import Any, Callable, Final, Iterable

import numpy as np
from pycommons.io.console import logger
from pycommons.io.csv import COMMENT_START, CSV_SEPARATOR
from pycommons.io.path import Path
from pycommons.strings.string_conv import num_to_str, str_to_num
from pycommons.types import type_error

from moptipy.api.logging import (
    KEY_ALGORITHM,
    KEY_GOAL_F,
    KEY_INSTANCE,
    KEY_RAND_SEED,
    PROGRESS_CURRENT_F,
    PROGRESS_FES,
    PROGRESS_TIME_MILLIS,
    SECTION_PROGRESS,
)
from moptipy.evaluation.base import (
    F_NAME_RAW,
    F_NAME_SCALED,
    KEY_ENCODING,
    KEY_OBJECTIVE_FUNCTION,
    TIME_UNIT_FES,
    TIME_UNIT_MILLIS,
    PerRunData,
    check_f_name,
    check_time_unit,
)
from moptipy.evaluation.log_parser import SetupAndStateParser
from moptipy.utils.logger import (
    KEY_VALUE_SEPARATOR,
)
from moptipy.utils.nputils import is_all_finite, is_np_float, is_np_int


@dataclass(frozen=True, init=False, order=False, eq=False)
class Progress(PerRunData):
    """An immutable record of progress information over a single run."""

    #: The time axis data.
    time: np.ndarray

    #: The unit of the time axis.
    time_unit: str

    #: The objective value data.
    f: np.ndarray

    #: the name of the objective value axis.
    f_name: str

    #: the standard value of the objective dimension.
    #: If :attr:`f_name` is `F_NAME_SCALED` or `F_NAME_NORMALIZED`.
    #: then this value has been used to normalize the data.
    f_standard: int | float | None

    def __init__(self,
                 algorithm: str,
                 instance: str,
                 objective: str,
                 encoding: str | None,
                 rand_seed: int,
                 time: np.ndarray,
                 time_unit: str,
                 f: np.ndarray,
                 f_name: str,
                 f_standard: int | float | None = None,
                 only_improvements: bool = True):
        """
        Create a consistent instance of :class:`EndResult`.

        :param algorithm: the algorithm name
        :param instance: the instance name
        :param objective: the name of the objective function
        :param encoding: the name of the encoding that was used, if any, or
            `None` if no encoding was used
        :param rand_seed: the random seed
        :param time: the time axis data
        :param time_unit: the unit of the time axis
        :param f: the objective value axis data
        :param f_name: the name of the objective value axis data
        :param f_standard: the value used to standardize of the objective
            value dimension
        :param only_improvements: enforce that f-values should be
            improving and time values increasing
        """
        super().__init__(algorithm, instance, objective, encoding, rand_seed)

        if not isinstance(time, np.ndarray):
            raise type_error(time, "time data", np.ndarray)
        time.flags.writeable = False
        if len(time.shape) != 1:
            raise ValueError("time array must be one-dimensional, but "
                             f"has shape {time.shape}.")
        if not is_np_int(time.dtype):
            raise TypeError("time data must be integer-valued, "
                            f"but has type {time.dtype}.")
        tl = time.size
        if tl <= 0:
            raise ValueError("time data must not be empty.")
        if tl > 1:
            if only_improvements:
                if np.any(time[1:] <= time[:-1]):
                    raise ValueError("time data must be strictly increasing,"
                                     f"but encountered {time}.")
            elif np.any(time[1:] < time[:-1]):
                raise ValueError("time data must be monotonously"
                                 f"increasing, but encountered {time}.")

        object.__setattr__(self, "time", time)
        object.__setattr__(self, "time_unit", check_time_unit(time_unit))

        mintime = 1 if time_unit == TIME_UNIT_FES else 0
        if np.any(time < mintime):
            raise ValueError(f"No time value can be less than {mintime} if"
                             f" time unit is {time_unit}.")

        if not isinstance(f, np.ndarray):
            raise type_error(f, "f data", np.ndarray)
        f.flags.writeable = False
        if len(f.shape) != 1:
            raise ValueError(
                f"f array must be one-dimensional, but has shape {f.shape}.")
        if is_np_float(f.dtype):
            if not is_all_finite(f):
                raise ValueError("f must be all finite.")
        elif not is_np_int(f.dtype):
            raise TypeError("f data must be integer- or float valued, but"
                            f" encountered an {type(f)} of {f.dtype}.")
        fl = f.size
        if fl <= 0:
            raise ValueError("f data must not be empty.")
        if fl != tl:
            raise ValueError(f"Length {fl} of f data and length {tl} of "
                             "time data must be the same.")
        if not isinstance(only_improvements, bool):
            raise type_error(only_improvements, "only_improvements", bool)
        if only_improvements and (fl > 1):
            if np.any(f[1:-1] >= f[:-2]):
                raise ValueError(
                    "f data must be strictly decreasing, with "
                    "only the entry being permitted as exception.")
            if f[-1] > f[-2]:
                raise ValueError(f"last f-value ({f[-1]}) cannot be greater"
                                 f"than second-to-last ({f[-2]}).")
        object.__setattr__(self, "f", f)
        object.__setattr__(self, "f_name", check_f_name(f_name))

        if (f_name != F_NAME_RAW) and (f_standard is None):
            raise ValueError(f"If f_name is {F_NAME_RAW}, "
                             f"then f_standard cannot be {f_standard}.")
        if f_standard is not None:
            if isinstance(f_standard, float):
                if not isfinite(f_standard):
                    raise ValueError(f"f_standard cannot be {f_standard}.")
            elif not isinstance(f_standard, int):
                raise type_error(f_standard, "f_standard", (int, float))
        object.__setattr__(self, "f_standard", f_standard)


def from_logs(path: str,
              consumer: Callable[[Progress], Any],
              time_unit: str = TIME_UNIT_FES,
              f_name: str = F_NAME_RAW,
              f_standard: dict[str, int | float] | None = None,
              only_improvements: bool = True) -> None:
    """
    Parse a given path and pass all progress data found to the consumer.

    If `path` identifies a file with suffix `.txt`, then this file is
    parsed. The appropriate :class:`Progress` is created and appended to
    the `collector`. If `path` identifies a directory, then this directory
    is parsed recursively for each log file found, one record is passed to
    the `consumer`. The `consumer` is simply a callable function. You could
    pass in the `append` method of a :class:`list`.

    :param path: the path to parse
    :param consumer: the consumer, can be the `append` method of a
        :class:`list`
    :param time_unit: the time unit
    :param f_name: the objective name
    :param f_standard: a dictionary mapping instances to standard values
    :param only_improvements: enforce that f-values should be improving and
        time values increasing
    """
    __InnerLogParser(time_unit, f_name, consumer, f_standard,
                     only_improvements).parse(path)


def to_csv(progress: Progress, file: str,
           put_header: bool = True) -> str:
    """
    Store a :class:`Progress` record in a CSV file.

    :param file: the file to generate
    :param put_header: should we put a header with meta-data?
    :return: the fully resolved file name
    """
    if not isinstance(progress, Progress):
        raise type_error(progress, "progress", Progress)
    if not isinstance(put_header, bool):
        raise type_error(put_header, "put_header", bool)
    path: Final[Path] = Path(file)
    logger(f"Writing progress object to CSV file {path!r}.")
    path.ensure_parent_dir_exists()

    with path.open_for_write() as out:
        sep: Final[str] = CSV_SEPARATOR
        write: Final[Callable[[str], int]] = out.write
        if put_header:
            kv: Final[str] = KEY_VALUE_SEPARATOR
            cmt: Final[str] = COMMENT_START
            write(f"{cmt} {KEY_ALGORITHM}{kv}{progress.algorithm}\n")
            write(f"{cmt} {KEY_INSTANCE}{kv}{progress.instance}\n")
            write(f"{cmt} {KEY_OBJECTIVE_FUNCTION}{kv}{progress.objective}\n")
            if progress.encoding is not None:
                write(f"{cmt} {KEY_ENCODING}{kv}{progress.objective}\n")
            write(f"{cmt} {KEY_RAND_SEED}{kv}{hex(progress.rand_seed)}\n")
            if progress.f_standard is not None:
                write(f"{cmt} {KEY_GOAL_F}{kv}{progress.f_standard}\n")
        write(f"{progress.time_unit}{sep}{progress.f_name}\n")
        for i, t in enumerate(progress.time):
            write(f"{t}{sep}{num_to_str(progress.f[i])}\n")

    logger(f"Done writing progress object to CSV file {path!r}.")

    path.enforce_file()
    return path


class __InnerLogParser(SetupAndStateParser):
    """The internal log parser class."""

    def __init__(self, time_unit: str, f_name: str,
                 consumer: Callable[[Progress], Any],
                 f_standard: dict[str, int | float] | None = None,
                 only_improvements: bool = True):
        """
        Create the internal log parser.

        :param time_unit: the time unit
        :param f_name: the objective name
        :param consumer: the consumer
        :param f_standard: a dictionary mapping instances to standard values
        :param only_improvements: enforce that f-values should be improving
            and time values increasing
        """
        super().__init__()
        if not callable(consumer):
            raise type_error(consumer, "consumer", call=True)
        self.__consumer: Final[Callable[[Progress], Any]] = consumer
        self.__time_unit = check_time_unit(time_unit)
        self.__f_name = check_f_name(f_name)
        self.__last_fe: int | None = None
        self.__t_collector: Final[list[int]] = []
        self.__f_collector: Final[list[int | float]] = []
        if not isinstance(only_improvements, bool):
            raise type_error(only_improvements, "only_improvements", bool)
        self.__only_improvements = only_improvements
        if (f_standard is not None) and (not isinstance(f_standard, dict)):
            raise type_error(f_standard, "f_standard", dict)
        self.__f_standard: Final[None | dict[str, int | float]] \
            = f_standard
        self.__state: int = 0

    def end_file(self) -> bool:
        if self.__state != 2:
            raise ValueError(
                "Illegal state, log file must have a "
                f"{SECTION_PROGRESS!r} section.")
        if not self.__f_collector:
            raise ValueError("f-collector cannot be empty.")
        if not self.__t_collector:
            raise ValueError("time-collector cannot be empty.")
        self.__state = 0
        return super().end_file()

    def process(self) -> None:
        f_standard: int | float | None = None
        if (self.__f_standard is not None) and \
                (self.instance in self.__f_standard):
            f_standard = self.__f_standard[self.instance]
        if f_standard is None:
            f_standard = self.goal_f
        if (self.__f_name != F_NAME_RAW) and (f_standard is None):
            raise ValueError(f"f_standard cannot be {f_standard} if f_name "
                             f"is {self.__f_name}.")

        tt = self.total_time_millis if (self.__time_unit == TIME_UNIT_MILLIS) \
            else self.total_fes
        if tt < self.__t_collector[-1]:
            raise ValueError(
                f"Last time units {tt} inconsistent with last"
                f"recorded time unit {self.__t_collector[-1]}.")
        if self.__last_fe < self.total_fes:
            if tt > self.__t_collector[-1]:
                self.__t_collector.append(tt)
                self.__f_collector.append(self.__f_collector[-1])
        elif self.__last_fe > self.total_fes:
            raise ValueError(
                f"Last FE {self.__last_fe} inconsistent with total number"
                f"{self.total_fes} of FEs.")

        ff: np.ndarray
        if self.__f_name == F_NAME_RAW:
            ff = np.array(self.__f_collector)
        elif self.__f_name == F_NAME_SCALED:
            ff = np.array([f / f_standard for f in self.__f_collector])
        else:
            ff = np.array([(f - f_standard) / f_standard
                           for f in self.__f_collector])
        self.__f_collector.clear()

        self.__consumer(Progress(self.algorithm,
                                 self.instance,
                                 self.objective,
                                 self.encoding,
                                 self.rand_seed,
                                 np.array(self.__t_collector),
                                 self.__time_unit,
                                 ff,
                                 self.__f_name,
                                 f_standard,
                                 self.__only_improvements))
        self.__t_collector.clear()
        self.__last_fe = None

    def start_section(self, title: str) -> bool:
        if title == SECTION_PROGRESS:
            if self.__state != 0:
                raise ValueError(f"Already did section {title}.")
            self.__state = 1
            return True
        return super().start_section(title)

    def needs_more_lines(self) -> bool:
        return (self.__state < 2) or super().needs_more_lines()

    def lines(self, lines: list[str]) -> bool:
        if not isinstance(lines, list):
            raise type_error(lines, "lines", list)
        if self.__state != 1:
            return super().lines(lines)
        n_rows = len(lines)
        if n_rows < 2:
            raise ValueError("lines must contain at least two elements,"
                             f"but contains {n_rows}.")

        columns = [c.strip() for c in lines[0].split(CSV_SEPARATOR)]
        n_cols = len(columns)
        if n_cols < 3:
            raise ValueError("There must be at least three columns, "
                             f"but found {n_cols} in {lines[0]!r}.")

        time_col_name: str = PROGRESS_TIME_MILLIS if \
            self.__time_unit == TIME_UNIT_MILLIS else PROGRESS_FES
        time_col_idx: int = -1
        f_col_idx: int = -1
        fe_col_idx: int = -1
        for idx, col in enumerate(columns):  # find the columns we
            if col == PROGRESS_FES:
                fe_col_idx = idx
            if col == time_col_name:
                if time_col_idx >= 0:
                    raise ValueError(f"Time column {time_col_name} "
                                     "appears twice.")
                time_col_idx = idx
            elif col == PROGRESS_CURRENT_F:
                if f_col_idx >= 0:
                    raise ValueError(
                        f"F column {PROGRESS_CURRENT_F} "
                        "appears twice.")
                f_col_idx = idx

        def aa(splt):  # noqa
            return splt[time_col_idx], splt[f_col_idx]

        time: Iterable[int]
        f: Iterable[Any]
        time, f = zip(*[[c.strip()
                         for c in aa(line.split(CSV_SEPARATOR))]
                        for line in lines[1:]], strict=True)
        time = [int(t) for t in time]
        f = [str_to_num(v) for v in f]
        if self.__only_improvements:
            biggest_t: int = -1
            best_f: int | float = inf
            for idx, t in enumerate(time):
                v = f[idx]
                if t > biggest_t:
                    if biggest_t >= 0:
                        self.__t_collector.append(biggest_t)
                        self.__f_collector.append(best_f)
                    best_f = v
                    biggest_t = t
                elif v < best_f:
                    best_f = v
            if biggest_t >= 0:
                self.__t_collector.append(biggest_t)
                self.__f_collector.append(best_f)
        else:
            self.__t_collector.extend(time)
            self.__f_collector.extend(f)

        self.__last_fe = int((lines[-1].split(CSV_SEPARATOR))[fe_col_idx])
        if self.__last_fe <= 0:
            raise ValueError(f"Last FE cannot be {self.__last_fe}.")

        self.__state = 2
        return self.needs_more_lines()



"""
Provides the CMA-ES Family Algorithms from the Library `cmaes`.

The Covariance Matrix Adaptation Evolutionary Strategy, CMA-ES for short, is a
very efficient optimization algorithm for small- and mid-scale and numerical/
continuous optimization problems.

Here, we wrap our `moptipy` API around the beautiful library `cmaes` by
Masashi Shibata and Masahiro Nomura at https://pypi.org/project/cmaes/. They
provide a CMA-ES implementation based on the ask-tell interface. In this
interface, you repeatedly query sample points in the search space from the
model and evaluate them. Then you feed back the points and their corresponding
objective values to the CMA-ES algorithm so that it can update its model. Then
the cycle is repeated.

1. Nikolaus Hansen and Andreas Ostermeier. A Completely Derandomized
   Self-Adaptation in Evolution Strategies. *Evolutionary Computation.*
   9(2):159-195. Summer 2001. https://dx.doi.org/10.1162/106365601750190398
2. Nikolaus Hansen. *The CMA Evolution Strategy: A Tutorial.*
   arXiv:1604.00772, 2016. https://arxiv.org/abs/1604.00772
3. Raymond Ros and Nikolaus Hansen. A Simple Modification in CMA-ES Achieving
   Linear Time and Space Complexity. In Günter Rudolph, Thomas Jansen, Nicola
   Beume, Simon Lucas, and Carlo Poloni, eds., Proceedings of the 10th
   International Conference on Parallel Problem Solving From Nature (PPSN X),
   September 13-17, 2008, Dortmund, Germany, pages 296-305. Volume 5199 of
   Lecture Notes in Computer Science. Berlin/Heidelberg, Germany: Springer.
   http://dx.doi.org/10.1007/978-3-540-87700-4_30
   https://hal.inria.fr/inria-00287367/document
4. Nikolaus Hansen. Benchmarking a BI-Population CMA-ES on the BBOB-2009
   Function Testbed. In Proceedings of the 11th Annual Conference Companion
   on Genetic and Evolutionary Computation Conference: Late Breaking Papers,
   July 8-12, 2009, Montreal, Québec, Canada, pages 2389-2396.
   New York, USA: ACM. http://dx.doi.org/10.1145/1570256.1570333
   https://hal.inria.fr/inria-00382093/document

- https://pypi.org/project/cmaes/
- https://github.com/CyberAgent/cmaes
"""

from typing import Callable, Final

import numpy as np
from cmaes import CMA, SepCMA  # type: ignore
from numpy.random import Generator
from pycommons.strings.string_conv import num_to_str
from pycommons.types import type_error

from moptipy.api.algorithm import Algorithm
from moptipy.api.process import Process
from moptipy.spaces.vectorspace import VectorSpace
from moptipy.utils.logger import CSV_SEPARATOR, KeyValueLogSection


def _run_cma(cma: SepCMA | CMA,
             f: Callable[[np.ndarray], int | float],
             should_terminate: Callable[[], bool],
             solutions: list[tuple[np.ndarray, int | float]],
             run_criterion: Callable[[], bool] = lambda: False) -> int:
    """
    Run a CMA implementation from the `cmaes` library.

    This is an internal core routine that translates the ask-tell interface
    of the algorithm implementations in the `cmaes` library into a simple
    loop.

    :param cma: the algorithm instance
    :param f: the objective function
    :param should_terminate: the termination criterion
    :param solutions: the internal list to store the solutions
    :param run_criterion: the stopper for a run
    :returns: the number of consumed FEs if the run was terminated by
        `run_criterion`, `-1` otherwise
    """
    fes: int = 0
    pop_size: Final[int] = cma.population_size

    # now we load a lot of fast call function pointers
    ask: Final[Callable[[], np.ndarray]] = cma.ask
    append: Final[Callable[[
        tuple[np.ndarray, int | float]], None]] = solutions.append
    tell: Final[Callable[
        [list[tuple[np.ndarray, float]]], None]] = cma.tell
    clear: Final[Callable[[], None]] = solutions.clear

    while True:  # the main loop
        clear()  # clear the ask/tell records
        for _ in range(pop_size):
            if should_terminate():  # budget over?
                return -1  # exit
            x: np.ndarray = ask()  # sample a point from CMA-ES
            value: int | float = f(x)  # compute its objective value
            append((x, value))  # store the point
            fes = fes + 1
        tell(solutions)  # feed all results back to the CMA
        if run_criterion():
            return fes


class CMAES(Algorithm):
    """
    A wrapper for the `CMA` algorithm from `cmaes`.

    1. Nikolaus Hansen and Andreas Ostermeier. A Completely Derandomized
       Self-Adaptation in Evolution Strategies. *Evolutionary Computation.*
       9(2):159-195. Summer 2001.
       https://dx.doi.org/10.1162/106365601750190398
    2. Nikolaus Hansen. *The CMA Evolution Strategy: A Tutorial.*
       arXiv:1604.00772, 2016. https://arxiv.org/abs/1604.00772
    """

    def __init__(self, space: VectorSpace) -> None:
        """
        Create the CMAES algorithm.

        :param space: the vector space
        """
        super().__init__()
        if not isinstance(space, VectorSpace):
            raise type_error(space, "space", VectorSpace)
        if space.dimension <= 1:
            raise ValueError("CMA-ES only works on at least two dimensions.")
        #: the vector space defining the dimensions and bounds
        self.space: Final[VectorSpace] = space

    def solve(self, process: Process) -> None:
        """
        Apply the bi-population CMA-ES to an optimization problem.

        :param process: the black-box process object
        """
        f: Final[Callable[[np.ndarray], int | float]] = \
            self.space.clipped(process.evaluate)  # the clipped objective
        should_terminate: Final[Callable[[], bool]] = \
            process.should_terminate  # the termination criterion

        lb: Final[np.ndarray] = self.space.lower_bound  # the upper bound
        ub: Final[np.ndarray] = self.space.upper_bound  # the lower bound
        mean: Final[np.ndarray] = 0.5 * (lb + ub)  # use center as mean value
        sigma: Final[float] = 0.2 * max(ub - lb)  # use a large initial sigma
        bounds: Final[np.ndarray] = \
            np.stack((lb, ub)).transpose()  # construct bounds

        # we create and directly run the CMA-ES algorithm
        _run_cma(CMA(mean=mean, sigma=sigma, bounds=bounds,
                     seed=process.get_random().integers(0, 4294967296)),
                 f, should_terminate, [])

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)  # log algorithm/operator
        self.space.log_bounds(logger)  # log bounds

    def __str__(self):
        """
        Get the name of this optimization algorithm.

        :retval "cmaes_cmaes": always
        """
        return "cmaes_cmaes"


class SepCMAES(CMAES):
    """
    The Separable CMA-ES based on Class `SepCMA` from Library `cmaes`.

    This is a variant of the CMA-ES where the covariance matrix is
    constrained to be diagonal. This means that there are fewer parameters to
    learn, so the learning rate for the covariance matrix can be increased.
    This algorithm is suitable if the problem is of larger scale, i.e., has
    a high dimension, in which case the pure CMA-ES may become rather slow in
    terms of its runtime consumption. Then, the loss of solution quality
    resulting from the underlying assumption that the objective function is
    separable is acceptable versus the gain in speed. By learning only the
    diagonals of the covariance matrix, the implicit assumption is that there
    are no mutual influences between the different decision variables. Of
    course, if the optimization problem is already of that nature, i.e.,
    separable, the algorithm will be faster than the normal CMA-ES at the same
    solution quality.

    1. Raymond Ros and Nikolaus Hansen. A Simple Modification in CMA-ES
       Achieving Linear Time and Space Complexity. In Günter Rudolph,
       Thomas Jansen, Nicola Beume, Simon Lucas, and Carlo Poloni, eds.,
       Proceedings of the 10th International Conference on Parallel
       Problem Solving From Nature (PPSN X), September 13-17, 2008,
       Dortmund, Germany, pages 296-305. Volume 5199 of Lecture Notes in
       Computer Science. Berlin/Heidelberg, Germany: Springer.
       http://dx.doi.org/10.1007/978-3-540-87700-4_30
       https://hal.inria.fr/inria-00287367/document
    """

    def solve(self, process: Process) -> None:
        """
        Apply the separable CMA-ES version to an optimization problem.

        :param process: the optimization problem to solve
        """
        f: Final[Callable[[np.ndarray], int | float]] = \
            self.space.clipped(process.evaluate)  # the clipped objective
        should_terminate: Final[Callable[[], bool]] = \
            process.should_terminate  # the termination criterion

        lb: Final[np.ndarray] = self.space.lower_bound  # the upper bound
        ub: Final[np.ndarray] = self.space.upper_bound  # the lower bound
        mean: Final[np.ndarray] = 0.5 * (lb + ub)  # use center as mean value
        sigma: Final[float] = 0.2 * max(ub - lb)  # use a large initial sigma
        bounds: Final[np.ndarray] = \
            np.stack((lb, ub)).transpose()  # construct bounds

        # we create and directly run the CMA-ES algorithm
        _run_cma(SepCMA(mean=mean, sigma=sigma, bounds=bounds,
                        seed=process.get_random().integers(0, 4294967296)),
                 f, should_terminate, [])

    def __str__(self):
        """
        Get the name of this optimization algorithm.

        :retval "sepCmaes_cmaes": always
        """
        return "sepCmaes_cmaes"


class BiPopCMAES(CMAES):
    """
    The bi-population CMA-ES based on Class `CMA` from Library `cmaes`.

    This algorithm combines two restart strategies for the normal CMA-ES under
    its hood. One where the population size increases exponentially and one
    where varying small population sizes are used.

    We here implement the bi-population CMA-ES algorithm in exactly the same
    way as the authors of the `cmaes` library do on
    https://pypi.org/project/cmaes/.

    1. Nikolaus Hansen. Benchmarking a BI-Population CMA-ES on the BBOB-2009
       Function Testbed. In Proceedings of the 11th Annual Conference
       Companion on Genetic and Evolutionary Computation Conference: Late
       Breaking Papers, July 8-12, 2009, Montreal, Québec, Canada,
       pages 2389-2396. New York, USA: ACM.
       http://dx.doi.org/10.1145/1570256.1570333
       https://hal.inria.fr/inria-00382093/document
    """

    def __init__(self, space: VectorSpace,
                 log_restarts: bool = False) -> None:
        """
        Create the CMAES algorithm.

        :param space: the vector space
        :param log_restarts: log the restart counters
        """
        super().__init__(space)
        if not isinstance(log_restarts, bool):
            raise type_error(log_restarts, "log_restarts", bool)
        #: should we log the FEs when the restarts happen or not?
        self.log_restarts: Final[bool] = log_restarts

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the algorithm to a logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)  # log algorithm/operator
        logger.key_value("logRestarts", self.log_restarts)

    def solve(self, process: Process) -> None:
        """
        Apply the external `cmaes` implementation to an optimization problem.

        :param process: the black-box process object
        """
        f: Final[Callable[[np.ndarray], int | float]] = \
            self.space.clipped(process.evaluate)  # the clipped objective
        should_terminate: Final[Callable[[], bool]] = \
            process.should_terminate  # the termination criterion
        # should we log the CMA-ES restart settings?
        restarts: Final[list[tuple[int, int, int, int, bool]] | None] = \
            [] if self.log_restarts and process.has_log() else None

        lb: Final[np.ndarray] = self.space.lower_bound  # the upper bound
        ub: Final[np.ndarray] = self.space.upper_bound  # the lower bound
        mean: Final[np.ndarray] = 0.5 * (lb + ub)  # use center as mean value
        sigma: Final[float] = 0.2 * max(ub - lb)  # use a large initial sigma
        bounds: Final[np.ndarray] = \
            np.stack((lb, ub)).transpose()  # construct bounds

        random: Generator = process.get_random()

        # create the initial CMA-ES setup
        seed: int = random.integers(0, 4294967296)
        cma = CMA(mean=mean, sigma=sigma, bounds=bounds, seed=seed)

        solutions: list[tuple[np.ndarray, int | float]] = []
        large_pop_restarts: int = 0  # the restarts with big population
        small_pop_fes: int = 0  # the FEs spent in the small population
        large_pop_fes: int = 0  # the FEs spent in the large population
        initial_pop_size: Final[int] = cma.population_size
        is_small_pop: bool = True  # are we in a small-population run?

        # The first run is with the "normal" population size. This is
        # the large population before the first doubling, but its FEs
        # count for the small population.
        while True:  # the main loop
            if restarts is not None:
                restarts.append((process.get_consumed_fes(),
                                 process.get_consumed_time_millis(),
                                 cma.population_size, seed, is_small_pop))
            fes = _run_cma(cma, f, should_terminate, solutions,
                           cma.should_stop)
            if fes < 0:  # this means that should_terminate became True
                break   # so we quit
            if is_small_pop:  # it was a small population so increment
                small_pop_fes += fes  # the small-population-FEs
            else:  # it was a large population, so increment the
                large_pop_fes += fes  # the large-population-FEs

            # We try to spend the same number FEs in small as in the large
            # population.
            is_small_pop = small_pop_fes < large_pop_fes

            if is_small_pop:  # create the small population
                pop_size_multiplier = 2 ** large_pop_restarts
                pop_size = max(1, int(
                    initial_pop_size * pop_size_multiplier ** (
                        random.uniform() ** 2)))
            else:  # else: create the large population
                large_pop_restarts = large_pop_restarts + 1
                pop_size = initial_pop_size * (2 ** large_pop_restarts)

            # Create the new CMA-ES instance.
            seed = random.integers(0, 4294967296)
            cma = CMA(mean=mean, sigma=sigma, bounds=bounds,
                      population_size=pop_size,
                      seed=seed)

        if restarts is not None:  # write the log section
            log: Final[list[str]] = [
                f"fes{CSV_SEPARATOR}timeMillis{CSV_SEPARATOR}popSize"
                f"{CSV_SEPARATOR}seed{CSV_SEPARATOR}isSmall"]
            for row in restarts:
                log.append(CSV_SEPARATOR.join(map(
                    num_to_str, (x for x in row))))
            del restarts
            process.add_log_section("CMA_RESTARTS", "\n".join(log))
            del log

    def __str__(self):
        """
        Get the name of this optimization algorithm.

        :retval "biPopCmaes_cmaes": always
        """
        return "biPopCmaes_cmaes"



"""Test stuff on ordered choices-based spaces."""

from typing import Any, Callable, Iterable, cast

import numpy as np
from numpy.random import Generator, default_rng
from pycommons.types import type_error

from moptipy.api.operators import Op0
from moptipy.spaces.ordered_choices import OrderedChoices
from moptipy.tests.op0 import validate_op0


def choices_for_tests(
        choice_filter: Callable[[OrderedChoices], bool] | None = None) \
        -> Iterable[OrderedChoices]:
    """
    Get a sequence of ordered choices for tests.

    :param choice_filter: an optional filter to sort out ordered choices we
        cannot use for testing
    :returns: the sequence of ordered choices
    """
    r = default_rng()
    pwrs: list[OrderedChoices] = [
        OrderedChoices([[1], [2]]),
        OrderedChoices([[1], [2], [3]]),
        OrderedChoices([[1], [2, 4], [3]]),
        OrderedChoices.signed_permutations(1),
        OrderedChoices.signed_permutations(2),
        OrderedChoices.signed_permutations(10)]

    for _i in range(4):
        done: set[int] = set()
        created: list[list[int]] = []
        choices: list[list[int]] = []
        while (len(choices) <= 0) or (len(done) < 2) or (r.integers(6) > 0):
            if r.integers(2) <= 0 < len(created):
                choices.append(created[r.integers(len(created))])
            else:
                picked: list[int] = []
                while (len(picked) <= 0) or (r.integers(6) > 0):
                    cc = r.integers(-100, 100)
                    while cc in done:
                        cc += 1
                    done.add(cc)
                    picked.append(int(cc))
                choices.append(picked)
                created.append(picked)
        pwrs.append(OrderedChoices(choices))

    if choice_filter is not None:
        if not callable(choice_filter):
            raise type_error(choice_filter, "choice_filter", None, call=True)
        pwrs = [p for p in pwrs if choice_filter(p)]
    r.shuffle(cast(list, pwrs))
    return pwrs


def make_choices_valid(choices: OrderedChoices) -> \
        Callable[[Generator, np.ndarray], np.ndarray]:
    """
    Create a function that can make ordered choices valid.

    :param choices: the ordered choices
    :returns: the function
    """

    def __make_valid(prnd: Generator, x: np.ndarray,
                     bb=choices.blueprint,
                     cc=choices.choices.__getitem__) -> np.ndarray:
        np.copyto(x, bb)
        prnd.shuffle(x)
        for i, e in enumerate(x):
            ff = cc(e)
            x[i] = ff[prnd.integers(len(ff))]
        return x

    return __make_valid


def validate_op0_on_1_choices(
        op0: Op0 | Callable[[OrderedChoices], Op0],
        search_space: OrderedChoices,
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, OrderedChoices], int] | None = None) -> None:
    """
    Validate the nullary operator on one `OrderedChoices` instance.

    :param op0: the operator or operator factory
    :param search_space: the search space
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    """
    args: dict[str, Any] = {
        "op0": op0(search_space) if callable(op0) else op0,
        "search_space": search_space,
        "make_search_space_element_valid":
            make_choices_valid(search_space),
    }
    if number_of_samples is not None:
        args["number_of_samples"] = number_of_samples
    if min_unique_samples is not None:
        args["min_unique_samples"] = min_unique_samples
    validate_op0(**args)


def validate_op0_on_choices(
        op0: Op0 | Callable[[OrderedChoices], Op0],
        number_of_samples: int | None = None,
        min_unique_samples: int | Callable[[
            int, OrderedChoices], int] | None = None,
        choice_filter: Callable[[OrderedChoices], bool] | None = None) \
        -> None:
    """
    Validate the nullary operator on several `OrderedChoices` instances.

    :param op0: the operator or operator factory
    :param number_of_samples: the optional number of samples
    :param min_unique_samples: the optional unique samples
    :param choice_filter: an optional filter to sort out ordered choices we
        cannot use for testing
    """
    for choices in choices_for_tests(choice_filter):
        validate_op0_on_1_choices(op0, choices, number_of_samples,
                                  min_unique_samples)
